"""
"Naive" PnLs with limited signal options and disregarding transaction costs.
"""

from dataclasses import dataclass
import warnings
from itertools import product
from typing import Dict, List, Optional, Tuple, Union
from numbers import Number

import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
import numpy as np
import pandas as pd
import seaborn as sns

from macrosynergy.management.simulate import make_qdf
from macrosynergy.management.utils import reduce_df, update_df
from macrosynergy.panel.make_zn_scores import make_zn_scores
from macrosynergy.signal import SignalReturnRelations


class NaivePnL:
    """
    Computes and collects illustrative PnLs with limited signal options and
    disregarding transaction costs.

    :param <pd.Dataframe> df: standardized DataFrame with the following necessary
        columns: 'cid', 'xcat', 'real_date' and 'value'.
    :param <str> ret: return category.
    :param <List[str]> sigs: signal categories. Able to pass in multiple possible signals
        to the Class' constructor and their respective vintages will be held on the
        instance's DataFrame. The signals can subsequently be referenced through the
        self.make_pnl() method which receives a single signal per call.
    :param <List[str]> cids: cross sections that are traded. Default is all in the
        dataframe.
    :param <str, List[str]> bms: list of benchmark tickers for which
        correlations are displayed against PnL strategies.
    :param <str> start: earliest date in ISO format. Default is None and earliest date
        in df is used.
    :param <str> end: latest date in ISO format. Default is None and latest date in df
        is used.
    :param <dict> blacklist: cross-sections with date ranges that should be excluded
        from the dataframe.

    """

    def __init__(
        self,
        df: pd.DataFrame,
        ret: str,
        sigs: List[str],
        cids: List[str] = None,
        bms: Union[str, List[str]] = None,
        start: str = None,
        end: str = None,
        blacklist: dict = None,
    ):
        df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")

        # Will host the benchmarks.
        self.dfd = df

        assert isinstance(ret, str), "The return category expects a single <str>."
        self.ret = ret
        xcats = [ret] + sigs

        cols = ["cid", "xcat", "real_date", "value"]
        # Potentially excludes the benchmarks but will be held on the instance level
        # through self.dfd.
        self.df, self.xcats, self.cids = reduce_df(
            df[cols], xcats, cids, start, end, blacklist, out_all=True
        )

        self.sigs = sigs

        ticker_func = lambda t: t[0] + "_" + t[1]
        self.tickers = list(map(ticker_func, product(self.cids, self.xcats)))

        self.df["real_date"] = pd.to_datetime(self.df["real_date"])

        # Data structure used to track all of the generated PnLs from make_pnl() method.
        self.pnl_names = []
        self.signal_df = {}
        self.black = blacklist

        self.bm_bool = isinstance(bms, (str, list))
        if self.bm_bool:
            bms = [bms] if isinstance(bms, str) else bms

            # Pass in the original DataFrame; negative signal will not have been applied
            # which will corrupt the use of the benchmark categories.
            bm_dict = self.add_bm(df=self.dfd, bms=bms, tickers=self.tickers)

            self._bm_dict = bm_dict

        self.pnl_params = {}

    def add_bm(self, df: pd.DataFrame, bms: List[str], tickers: List[str]):
        """
        Returns a dictionary with benchmark return series.

        :param <pd.DataFrame> df: aggregate DataFrame passed into the Class.
        :param <List[str]> bms: benchmark return tickers.
        :param <List[str]> tickers: the available tickers held in the reduced DataFrame.
            The reduced DataFrame consists exclusively of the signal & return categories.
        """

        bm_dict = {}

        for bm in bms:
            # Accounts for appending "_NEG" to the ticker.
            bm_s = bm.split("_", maxsplit=1)
            cid = bm_s[0]
            xcat = bm_s[1]
            dfa = df[(df["cid"] == cid) & (df["xcat"] == xcat)]

            if dfa.shape[0] == 0:
                print(f"{bm} has no observations in the DataFrame.")
            else:
                df_single_bm = dfa.pivot(
                    index="real_date", columns="xcat", values="value"
                )
                df_single_bm.columns = [bm]
                bm_dict[bm] = df_single_bm
                if bm not in tickers:
                    self.df = update_df(self.df, dfa)

        return bm_dict

    @staticmethod
    def _make_signal(
        dfx: pd.DataFrame,
        sig: str,
        sig_op: str = "zn_score_pan",
        min_obs: int = 252,
        iis: bool = True,
        sequential: bool = True,
        neutral: str = "zero",
        thresh: float = None,
    ):
        """
        Helper function used to produce the raw signal that forms the basis for
        positioning.

        :param <pd.DataFrame> dfx: DataFrame defined over the return & signal category.
        :param <str> sig: name of the raw signal.
        :param <str> sig_op: signal transformation.
        :param <int> min_obs: the minimum number of observations required to calculate
            zn_scores. Default is 252.
        :param <bool> iis: if True (default) zn-scores are also calculated for the initial
            sample period defined by min_obs, on an in-sample basis, to avoid losing
            history.
        :param <bool> sequential: if True (default) score parameters are estimated
            sequentially with concurrently available information only.
        :param <str> neutral: method to determine neutral level.
        :param <float> thresh: threshold value beyond which scores are winsorized,

        """

        if sig_op == "binary":
            dfw = dfx.pivot(index=["cid", "real_date"], columns="xcat", values="value")
            dfw["psig"] = np.sign(dfw[sig])
        elif sig_op == "raw":
            dfw = dfx.pivot(index=["cid", "real_date"], columns="xcat", values="value")
            dfw["psig"] = dfw[sig]
        else:
            panw = 1 if sig_op == "zn_score_pan" else 0
            # The re-estimation frequency for the neutral level and standard deviation
            # will be the same as the re-balancing frequency. For instance, if the
            # neutral level is computed weekly, a material change in the signal will only
            # manifest along a similar timeline. Therefore, re-estimation and
            # re-balancing frequencies match.
            df_ms = make_zn_scores(
                dfx,
                xcat=sig,
                neutral=neutral,
                pan_weight=panw,
                sequential=sequential,
                min_obs=min_obs,
                iis=iis,
                thresh=thresh,
            )

            df_ms = df_ms.drop("xcat", axis=1)
            df_ms["xcat"] = "psig"

            dfx_concat = pd.concat([dfx, df_ms])
            dfw = dfx_concat.pivot(
                index=["cid", "real_date"], columns="xcat", values="value"
            )

        # Reconstruct the DataFrame to recognise the signal's start date for each
        # individual cross-section
        dfw_list = []
        for c, cid_df in dfw.groupby(level=0):
            first_date = cid_df.loc[:, "psig"].first_valid_index()
            cid_df = cid_df.loc[first_date:, :]
            dfw_list.append(cid_df)
        return pd.concat(dfw_list)

    @staticmethod
    def rebalancing(dfw: pd.DataFrame, rebal_freq: str = "daily", rebal_slip=0):
        """
        The signals are calculated daily and for each individual cross-section defined in
        the panel. However, re-balancing a position can occur more infrequently than
        daily. Therefore, produce the re-balancing values according to the more
        infrequent timeline (weekly or monthly).

        :param <pd.Dataframe> dfw: DataFrame with each category represented by a column
            and the daily signal is also included with the column name 'psig'.
        :param <str> rebal_freq: re-balancing frequency for positions according to signal
            must be one of 'daily' (default), 'weekly' or 'monthly'.
        :param <str> rebal_slip: re-balancing slippage in days.

        :return <pd.Series>: will return a pd.Series containing the associated signals
            according to the re-balancing frequency.
        """

        # The re-balancing days are the first of the respective time-periods because of
        # the shift forward by one day applied earlier in the code. Therefore, only
        # concerned with the minimum date of each re-balance period.
        dfw["year"] = dfw["real_date"].dt.year
        if rebal_freq == "monthly":
            dfw["month"] = dfw["real_date"].dt.month
            rebal_dates = dfw.groupby(["cid", "year", "month"])["real_date"].min()
        elif rebal_freq == "weekly":
            dfw["week"] = dfw["real_date"].apply(lambda x: x.week)
            rebal_dates = dfw.groupby(["cid", "year", "week"])["real_date"].min()

        # Convert the index, 'cid', to a formal column aligned to the re-balancing dates.
        r_dates_df = rebal_dates.reset_index(level=0)
        r_dates_df.reset_index(drop=True, inplace=True)
        dfw = dfw[["real_date", "psig", "cid"]]

        # Isolate the required signals on the re-balancing dates. Only concerned with the
        # respective signal on the re-balancing date. However, the produced DataFrame
        # will only be defined over the re-balancing dates. Therefore, merge the
        # aforementioned DataFrame with the original DataFrame such that all business
        # days are included. The intermediary dates, dates between re-balancing dates,
        # will initially be populated by NA values. To ensure the signal is used for the
        # duration between re-balancing dates, forward fill the computed signal over the
        # associated dates.

        # The signal is computed for each individual cross-section. Therefore, merge on
        # the real_date and the cross-section.
        rebal_merge = r_dates_df.merge(dfw, how="left", on=["real_date", "cid"])
        # Re-establish the daily date series index where the intermediary dates, between
        # the re-balancing dates, will be populated using a forward fill.
        rebal_merge = dfw[["real_date", "cid"]].merge(
            rebal_merge, how="left", on=["real_date", "cid"]
        )
        rebal_merge["psig"] = rebal_merge["psig"].ffill().shift(rebal_slip)
        rebal_merge = rebal_merge.sort_values(["cid", "real_date"])

        rebal_merge = rebal_merge.set_index("real_date")
        sig_series = rebal_merge.drop(["cid"], axis=1)

        return sig_series

    def make_pnl(
        self,
        sig: str,
        sig_op: str = "zn_score_pan",
        sig_add: float = 0,
        sig_neg: bool = False,
        pnl_name: Optional[str] = None,
        rebal_freq: str = "daily",
        rebal_slip: int = 0,
        vol_scale: Optional[float] = None,
        leverage: float = 1.0,
        min_obs: int = 261,
        iis: bool = True,
        sequential: bool = True,
        neutral: str = "zero",
        thresh: float = None,
    ):
        """
        Calculate daily PnL and add to class instance.

        :param <str> sig: name of raw signal that is basis for positioning. The signal
            is assumed to be recorded at the end of the day prior to position taking.
        :param <str> sig_op: signal transformation options; must be one of
            'zn_score_pan', 'zn_score_cs', 'binary', or 'raw'. The default is 'zn_score_pan'.
            'zn_score_pan': transforms raw signals into z-scores around zero value
            based on the whole panel. The neutral level & standard deviation will use the
            cross-section of panels.
            'zn_score_cs': transforms signals to z-scores around zero based on
            cross-section alone.
            'binary': transforms signals into uniform long/shorts (1/-1) across all
            sections.
            N.B.: zn-score here means standardized score with zero being the natural
            neutral level and standardization through division by mean absolute value.
        :param <float> sig_add: add a constant to the signal after initial transformation.
            This allows to give PnLs a long or short bias relative to the signal
            score. Default is 0.
        :param <str> sig_neg: if True the PnL is based on the negative value of the
            transformed signal. Default is False.
        :param <str> pnl_name: name of the PnL to be generated and stored.
            Default is None, i.e. a default name is given. The default name will be:
            'PNL_<signal name>[_<NEG>]', with the last part added if sig_neg has been
            set to True.
            Previously calculated PnLs of the same name will be overwritten. This means
            that if a set of PnLs are to be compared, each PnL requires a distinct name.
        :param <str> rebal_freq: re-balancing frequency for positions according to signal
            must be one of 'daily' (default), 'weekly' or 'monthly'. The re-balancing is
            only concerned with the signal value on the re-balancing date which is
            delimited by the frequency chosen.
            Additionally, the re-balancing frequency will be applied to make_zn_scores()
            if used as the method to produce the raw signals.
        :param <str> rebal_slip: re-balancing slippage in days. Default is 1 which
            means that it takes one day to re-balance the position and that the new
            positions produce PnL from the second day after the signal has been recorded.
        :param <bool> vol_scale: ex-post scaling of PnL to annualized volatility given.
            This is for comparative visualization and not out-of-sample. Default is none.
        :param <float> leverage: leverage applied to the raw signal when a `vol_scale` is
            not defined. Default is 1.0, i.e., position size is 1 or 100% of implied risk
            capital.
        :param <int> min_obs: the minimum number of observations required to calculate
            zn_scores. Default is 252.
        :param <bool> iis: if True (default) zn-scores are also calculated for the initial
            sample period defined by min_obs, on an in-sample basis, to avoid losing
            history.
        :param <bool> sequential: if True (default) score parameters (neutral level and
            standard deviations) are estimated sequentially with concurrently available
            information only.
        :param <str> neutral: method to determine neutral level. Default is 'zero'.
            Alternatives are 'mean' and "median".
        :param <float> thresh: threshold value beyond which scores are winsorized,
            i.e. contained at that threshold. Therefore, the threshold is the maximum
            absolute score value that the function is allowed to produce. The minimum
            threshold is one standard deviation. Default is no threshold.

        """
        for varx, name, typex in (
            (sig, "sig", str),
            (sig_op, "sig_op", str),
            (sig_add, "sig_add", (Number)), # testing for number instead of (float, int)
            (sig_neg, "sig_neg", bool),
            (pnl_name, "pnl_name", (str, type(None))),
            (rebal_freq, "rebal_freq", str),
            (rebal_slip, "rebal_slip", int),
            (vol_scale, "vol_scale", (Number, type(None))),
            (leverage, "leverage", (Number)),
            (min_obs, "min_obs", int),
            (iis, "iis", bool),
            (sequential, "sequential", bool),
            (neutral, "neutral", str),
            (thresh, "thresh", (Number, type(None))),
        ):
            if not isinstance(varx, typex):
                raise TypeError(f"{name} must be a {typex}.")

        error_sig = (
            f"Signal category missing from the options defined on the class: "
            f"{self.sigs}. "
        )
        if sig not in self.sigs:
            raise ValueError(error_sig)

        sig_options = ["zn_score_pan", "zn_score_cs", "binary", "raw"]
        error_sig_method = (
            f"The signal transformation method, {sig_op}, is not one of "
            f"the options specified: {sig_options}."
        )
        if sig_op not in sig_options:
            raise ValueError(error_sig_method)

        freq_params = ["daily", "weekly", "monthly"]
        freq_error = f"Re-balancing frequency must be one of: {freq_params}."
        if rebal_freq not in freq_params:
            raise ValueError(freq_error)

        if thresh is not None and thresh < 1:
            raise ValueError("thresh must be greater than or equal to one.")

        err_lev = "`leverage` must be a numerical value greater than 0."
        if leverage <= 0:
            raise ValueError(err_lev)
        err_vol = "`vol_scale` must be a numerical value greater than 0."
        if vol_scale is not None and (vol_scale <= 0):
            raise ValueError(err_vol)

        # B. Extract DataFrame of exclusively return and signal categories in time series
        # format.
        dfx = self.df[self.df["xcat"].isin([self.ret, sig])]

        dfw = self._make_signal(
            dfx=dfx,
            sig=sig,
            sig_op=sig_op,
            min_obs=min_obs,
            iis=iis,
            sequential=sequential,
            neutral=neutral,
            thresh=thresh,
        )

        if sig_neg:
            dfw["psig"] *= -1
            neg = "_NEG"
        else:
            neg = ""

        dfw["psig"] += sig_add

        self._winsorize(df=dfw["psig"], thresh=thresh)

        # Multi-index DataFrame with a natural minimum lag applied.
        dfw["psig"] = dfw["psig"].groupby(level=0).shift(1)
        dfw.reset_index(inplace=True)
        dfw = dfw.rename_axis(None, axis=1)

        dfw = dfw.sort_values(["cid", "real_date"])

        if rebal_freq != "daily":
            sig_series = self.rebalancing(
                dfw=dfw, rebal_freq=rebal_freq, rebal_slip=rebal_slip
            )
            dfw["sig"] = np.squeeze(sig_series.to_numpy())
        else:
            dfw = dfw.rename({"psig": "sig"}, axis=1)

        # The signals are generated across the panel.
        dfw["value"] = dfw[self.ret] * dfw["sig"]

        df_pnl = dfw.loc[:, ["cid", "real_date", "value"]]

        # Compute the return across the panel. The returns are still computed daily
        # regardless of the re-balancing frequency potentially occurring weekly or
        # monthly.
        df_pnl_all = df_pnl.groupby(["real_date"]).sum(numeric_only=True)
        df_pnl_all = df_pnl_all[df_pnl_all["value"].cumsum() != 0]
        # Returns are computed for each cross-section and across the panel.
        df_pnl_all["cid"] = "ALL"
        df_pnl_all = df_pnl_all.reset_index()[df_pnl.columns]
        # Will be inclusive of each individual cross-section's signal-adjusted return and
        # the aggregated panel return.
        df_pnl = pd.concat([df_pnl, df_pnl_all])

        if vol_scale is not None:
            leverage = vol_scale * (df_pnl_all["value"].std() * np.sqrt(261)) ** (-1)

        assert isinstance(leverage, (float, int)), err_lev  # sanity check
        df_pnl["value"] = df_pnl["value"] * leverage

        pnn = ("PNL_" + sig + neg) if pnl_name is None else pnl_name
        # Populating the signal dictionary is required for the display methods:
        self.signal_df[pnn] = dfw.loc[:, ["cid", "real_date", "sig"]]

        df_pnl["xcat"] = pnn
        if pnn in self.pnl_names:
            self.df = self.df[~(self.df["xcat"] == pnn)]
        else:
            self.pnl_names = self.pnl_names + [pnn]

        agg_df = pd.concat([self.df, df_pnl[self.df.columns]])
        self.df = agg_df.reset_index(drop=True)

        self.pnl_params[pnn] = PnLParams(
            pnl_name=pnn,
            signal=sig,
            sig_op=sig_op,
            sig_add=sig_add,
            sig_neg=sig_neg,
            rebal_freq=rebal_freq,
            rebal_slip=rebal_slip,
            vol_scale=vol_scale,
            neutral=neutral,
            thresh=thresh,
        )

    def make_long_pnl(
        self,
        vol_scale: Optional[float] = None,
        label: Optional[str] = None,
        leverage: float = 1.0,
    ):
        """
        The long-only returns will be computed which act as a basis for comparison
        against the signal-adjusted returns. Will take a long-only position in the
        category passed to the parameter 'self.ret'.

        :param <bool> vol_scale: ex-post scaling of PnL to annualized volatility given.
            This is for comparative visualization and not out-of-sample, and is applied
            to the long-only position. Default is None.
        :param <str> label: associated label that will be mapped to the long-only
            DataFrame. The label will be used in the plotting graphic for plot_pnls().
            If a label is not defined, the default will be the name of the return
            category.
        :param <float> leverage: leverage applied to the raw signal when a `vol_scale` is
            not defined. Default is 1.0, i.e., position size is 1 or 100% of implied risk
            capital.
        """

        if vol_scale is not None:
            vol_err = (
                "The parameter `vol_scale` must be a numerical value greater than 0."
            )
            if not isinstance(vol_scale, (float, int)):
                raise TypeError(vol_err)
            elif vol_scale <= 0:
                raise ValueError(vol_err)

        else:
            err_lev = "`leverage` must be a numerical value greater than 0."
            if not isinstance(leverage, (float, int)):
                raise TypeError(err_lev)
            elif leverage <= 0:
                raise ValueError(err_lev)

        if label is None:
            label = self.ret

        dfx = self.df[self.df["xcat"].isin([self.ret])]

        df_long = self.long_only_pnl(
            dfw=dfx, vol_scale=vol_scale, label=label, leverage=leverage
        )

        self.df: pd.DataFrame = pd.concat([self.df, df_long])

        if label not in self.pnl_names:
            self.pnl_names = self.pnl_names + [label]

        self.df = self.df.reset_index(drop=True)

    @staticmethod
    def long_only_pnl(
        dfw: pd.DataFrame,
        vol_scale: float = None,
        label: str = None,
        leverage: float = 1.0,
    ):
        """
        Method used to compute the PnL accrued from simply taking a long-only position in
        the category, 'self.ret'. The returns from the category are not predicated on any
        exogenous signal.

        :param <pd.DataFrame> dfw:
        :param <bool> vol_scale: ex-post scaling of PnL to annualized volatility given.
            This is for comparative visualization and not out-of-sample. Default is none.
        :param <str> label: associated label that will be mapped to the long-only
            DataFrame.
        :param <float> leverage: leverage applied to the raw signal when a `vol_scale` is
            not defined. Default is 1.0, i.e., position size is 1 or 100% of implied risk
            capital.

        :return <pd.DataFrame> panel_pnl: standardised dataframe containing exclusively
            the return category, and the long-only panel return.
        """
        lev_err = "`leverage` must be a numerical value greater than 0."

        dfw_long = dfw.reset_index(drop=True)

        panel_pnl = dfw_long.groupby(["real_date"]).sum(numeric_only=True)
        panel_pnl = panel_pnl.reset_index(level=0)
        panel_pnl["cid"] = "ALL"
        panel_pnl["xcat"] = label

        if vol_scale is not None:
            leverage = vol_scale * (panel_pnl["value"].std() * np.sqrt(261)) ** (-1)

        if not isinstance(leverage, Number) or leverage <= 0:
            raise TypeError(lev_err)
        panel_pnl["value"] = panel_pnl["value"] * leverage

        return panel_pnl[["cid", "xcat", "real_date", "value"]]

    def plot_pnls(
        self,
        pnl_cats: List[str] = None,
        pnl_cids: List[str] = ["ALL"],
        start: str = None,
        end: str = None,
        facet: bool = False,
        ncol: int = 3,
        same_y: bool = True,
        title: str = "Cumulative Naive PnL",
        title_fontsize: int = 20,
        xcat_labels: Union[List[str], dict] = None,
        xlab: str = "",
        ylab: str = "% of risk capital, no compounding",
        share_axis_labels: bool = True,
        figsize: Tuple = (12, 7),
        aspect: float = 1.7,
        height: float = 3,
        label_adj: float = 0.05,
        title_adj: float = 0.95,
        y_label_adj: float = 0.95,
    ) -> None:
        """
        Plot line chart of cumulative PnLs, single PnL, multiple PnL types per
        cross section, or multiple cross sections per PnL type.

        :param <List[str]> pnl_cats: list of PnL categories that should be plotted.
        :param <List[str]> pnl_cids: list of cross sections to be plotted;
            default is 'ALL' (global PnL).
            Note: one can only have multiple PnL categories or multiple cross sections,
            not both.
        :param <str> start: earliest date in ISO format. Default is None and earliest
            date in df is used.
        :param <str> end: latest date in ISO format. Default is None and latest date
            in df is used.
        :param <bool> facet: parameter to control whether each PnL series is plotted on
            its own respective grid using Seaborn's FacetGrid. Default is False and all
            series will be plotted in the same graph.
        :param <int> ncol: number of columns in facet grid. Default is 3. If the total
            number of PnLs is less than ncol, the number of columns will be adjusted on
            runtime.
        :param <bool> same_y: if True (default) all plots in facet grid share same y axis.
        :param <str> title: allows entering text for a custom chart header.
        :param <int> title_fontsize: font size for the title. Default is 20.
        :param <List[str]> xcat_labels: custom labels to be used for the PnLs.
        :param <str> xlab: label for x-axis of the plot (or subplots if faceted),
            default is None (empty string)..
        :param <str> ylab: label for y-axis of the plot (or subplots if faceted),
            default is '% of risk capital, no compounding'.
        :param <bool> share_axis_labels: if True (default) the axis labels are shared by
            all subplots in the facet grid.
        :param <tuple> figsize: tuple of plot width and height. Default is (12 , 7).
        :param <float> aspect: width-height ratio for plots in facet. Default is 1.7.
        :param <float> height: height of plots in facet. Default is 3.
        :param <float> label_adj: parameter that sets bottom of figure to fit the label.
            Default is 0.05.
        :param <float> title_adj: parameter that sets top of figure to accommodate title.
            Default is 0.95.
        :param <float> y_label_adj: parameter that sets left of figure to fit the y-label.
        """

        if pnl_cats is None:
            pnl_cats = self.pnl_names
        else:
            pnl_cats_error = (
                f"List of PnL categories expected - received {type(pnl_cats)}."
            )
            if not isinstance(pnl_cats, list):
                raise TypeError(pnl_cats_error)

            pnl_cats_copy = pnl_cats.copy()
            pnl_cats = [pnl for pnl in pnl_cats if pnl in self.pnl_names]

            dif = set(pnl_cats_copy).difference(set(pnl_cats))
            if dif:
                print(
                    f"The PnL(s) requested, {dif}, have not been defined on the "
                    f"Class. The defined PnL(s) are {self.pnl_names}."
                )
            elif len(pnl_cats) == 0:
                raise ValueError(
                    "There are not any valid PnL(s) to display given the " "request."
                )

        error_message = "Either pnl_cats or pnl_cids must be a list of length 1"
        if not ((len(pnl_cats) == 1) | (len(pnl_cids) == 1)):
            raise ValueError(error_message)

        # adjust ncols of the facetgrid if necessary
        if max([len(pnl_cats), len(pnl_cids)]) < ncol:
            ncol = max([len(pnl_cats), len(pnl_cids)])

        dfx = reduce_df(
            self.df, pnl_cats, pnl_cids, start, end, self.black, out_all=False
        )

        if max([len(pnl_cats), len(pnl_cids)]) < ncol:
            ncol = max([len(pnl_cats), len(pnl_cids)])

        error_message = (
            "The number of custom labels must match the defined number of "
            "categories in pnl_cats."
        )

        if isinstance(xcat_labels, dict) or xcat_labels is None:
            if xcat_labels is None:
                xcat_labels = pnl_cats.copy()
            else:
                if not len(xcat_labels) == len(pnl_cats):
                    raise ValueError(error_message)
                xcat_labels = [xcat_labels[pnl] for pnl in pnl_cats]

        elif isinstance(xcat_labels, list) and all(
            isinstance(item, str) for item in xcat_labels
        ):
            warnings.warn(
                "xcat_labels should be a dictionary with keys as pnl_cats and values as "
                "the custom labels. This will be enforced in a future version.",
            )
            if len(xcat_labels) != len(pnl_cats):
                raise ValueError(error_message)
        else:
            raise TypeError(
                "xcat_labels should be a dictionary with keys as pnl_cats and values as "
                "the custom labels."
            )

        no_cids = len(pnl_cids)

        sns.set_theme(
            style="whitegrid", palette="colorblind", rc={"figure.figsize": figsize}
        )

        if no_cids == 1:
            plot_by = "xcat"
            col_order = pnl_cats
            labels = xcat_labels
            legend_title = "PnL Category(s)"
        else:
            plot_by = "cid"
            col_order = pnl_cids
            if xcat_labels is not None:
                labels = pnl_cids.copy()
            legend_title = "Cross Section(s)"

        dfx["cum_value"] = dfx.groupby(plot_by).cumsum(numeric_only=True)

        if facet:
            fg = sns.FacetGrid(
                data=dfx,
                col=plot_by,
                col_wrap=ncol,
                sharey=same_y,
                aspect=aspect,
                height=height,
                col_order=col_order,
                legend_out=True,
            )
            fg.fig.suptitle(
                title,
                fontsize=title_fontsize,
            )

            fg.fig.subplots_adjust(top=title_adj, bottom=label_adj, left=y_label_adj)

            fg.map_dataframe(
                sns.lineplot,
                x="real_date",
                y="cum_value",
                hue=plot_by,
                hue_order=col_order,
                estimator=None,
                lw=1,
            )
            for ix, ax in enumerate(fg.axes.flat):
                ax.axhline(y=0, color="black", linestyle="--", linewidth=1)
                if no_cids == 1:
                    ax.set_title(xcat_labels[ix])

            if no_cids > 1:
                fg.set_titles(row_template="", col_template="{col_name}")

            if share_axis_labels:
                fg.set_axis_labels("", "")
                fg.fig.supxlabel(xlab)
                fg.fig.supylabel(ylab)
            else:
                fg.set_axis_labels(xlab, ylab)

        else:
            fg = sns.lineplot(
                data=dfx,
                x="real_date",
                y="cum_value",
                hue=plot_by,
                hue_order=col_order,
                estimator=None,
                lw=1,
            )
            leg = fg.axes.get_legend()
            plt.title(title, fontsize=title_fontsize)
            plt.legend(
                labels=labels,
                title=legend_title,
            )
            plt.xlabel(xlab)
            plt.ylabel(ylab)

        if no_cids == 1:
            if facet:
                labels = labels[::-1]
        else:
            labels = labels[::-1]

        plt.axhline(y=0, color="black", linestyle="--", lw=1)
        plt.show()

    def signal_heatmap(
        self,
        pnl_name: str,
        pnl_cids: List[str] = None,
        start: str = None,
        end: str = None,
        freq: str = "m",
        title: str = "Average applied signal values",
        x_label: str = "",
        y_label: str = "",
        figsize: Optional[Tuple[float, float]] = None,
    ):
        """
        Display heatmap of signals across times and cross-sections.

        :param <str> pnl_name: name of naive PnL whose signals are displayed.
            N.B.: Signal is here is the value that actually determines
            the concurrent PnL.
        :param <List[str]> pnl_cids: cross-sections. Default is all available.
        :param <str> start: earliest date in ISO format. Default is None and earliest
            date in df is used.
        :param <str> end: latest date in ISO format. Default is None and latest date
            in df is used.
        :param <str> freq: frequency for which signal average is displayed.
            Default is monthly ('m'). The only alternative is quarterly ('q').
        :param <str> title: allows entering text for a custom chart header.
        :param <str> x_label: label for the x-axis. Default is None.
        :param <str> y_label: label for the y-axis. Default is None.
        :param <(float, float)> figsize: width and height in inches.
            Default is (14, number of cross sections).
        """

        if not isinstance(pnl_name, str):
            raise TypeError("The method expects to receive a single PnL name.")
        error_cats = (
            f"The PnL passed to 'pnl_name' parameter is not defined. The "
            f"possible options are {self.pnl_names}."
        )
        if pnl_name not in self.pnl_names:
            raise ValueError(error_cats)

        error_time = "Defined time-period must be monthly ('m') or quarterly ('q')"
        if not isinstance(freq, str) or freq not in ["m", "q"]:
            raise ValueError(error_time)

        error_cids = (
            f"Cross-sections not available. Available cids are:" f"{self.cids}."
        )

        if pnl_cids is None:
            pnl_cids = self.cids
        else:
            if not set(pnl_cids) <= set(self.cids):
                raise ValueError(error_cids)

        for vname, lbl in zip(["x_label", "y_label"], [x_label, y_label]):
            if not isinstance(lbl, str):
                raise TypeError(f"<str> expected for `{vname}` - received {type(lbl)}.")

        dfx: pd.DataFrame = self.signal_df[pnl_name]
        dfw: pd.DataFrame = dfx.pivot(index="real_date", columns="cid", values="sig")
        dfw: pd.DataFrame = dfw[pnl_cids]

        if start is None:
            start = dfw.index[0]
        elif end is None:
            end = dfw.index[-1]

        dfw = dfw.truncate(before=start, after=end)

        dfw = dfw.resample(freq).mean()

        if figsize is None:
            figsize = (14, len(pnl_cids))

        fig, ax = plt.subplots(figsize=figsize)
        dfw = dfw.transpose()
        dfw.columns = [str(d.strftime("%d-%m-%Y")) for d in dfw.columns]
        sns.heatmap(dfw, cmap="vlag_r", center=0)

        ax.set(xlabel=x_label, ylabel=y_label)
        ax.set_yticklabels(ax.get_yticklabels(), rotation=0)
        ax.set_title(title, fontsize=14)

        plt.show()

    def agg_signal_bars(
        self,
        pnl_name: str,
        freq: str = "m",
        metric: str = "direction",
        title: str = None,
        y_label: str = "Sum of Std. across the Panel",
    ):
        """
        Display aggregate signal strength and - potentially - direction.

        :param <str> pnl_name: name of the PnL whose signal is to be visualized.
            N.B.: The referenced signal corresponds to the series that determines the
            concurrent PnL.
        :param <str> freq: frequency at which the signal is visualized. Default is
            monthly ('m'). The alternative is quarterly ('q').
        :param <str> metric: the type of signal value. Default is "direction".
            Alternative is "strength".
        :param <str> title: allows entering text for a custom chart header. Default will
            be "Directional Bar Chart of <pnl_name>.".
        :param <str> y_label: label for the y-axis. Default is the sum of standard
            deviations across the panel corresponding to the default signal
            transformation: 'zn_score_pan'.

        """

        assert isinstance(pnl_name, str), (
            "The method expects to receive a single " "PnL name."
        )
        error_cats = (
            f"The PnL passed to 'pnl_name' parameter is not defined. The "
            f"possible options are {self.pnl_names}."
        )
        assert pnl_name in self.pnl_names, error_cats

        error_time = "Defined time-period must either be monthly, m, or quarterly, q."
        assert isinstance(freq, str) and freq in ["m", "q"], error_time

        metric_error = "The metric must either be 'direction' or 'strength'."
        assert metric in ["direction", "strength"], metric_error

        if title is None:
            title = f"Directional Bar Chart of {pnl_name}."

        dfx: pd.DataFrame = self.signal_df[pnl_name]
        dfw: pd.DataFrame = dfx.pivot(index="real_date", columns="cid", values="sig")

        # The method allows for visually understanding the overall direction of the
        # aggregate signal but also gaining insight into the proportional exposure to the
        # respective signal by measuring the absolute value, the size of the signal.
        # Is the PnL value generated by large returns or a large signal, position ?
        if metric == "strength":
            dfw = dfw.abs()

        dfw = dfw.resample(freq).mean()
        # Sum across the timestamps to compute the aggregate signal according to the
        # down-sampling frequency.
        df_s = dfw.sum(axis=1)
        index = np.array(df_s.index)
        df_signal = pd.DataFrame(
            data=df_s.to_numpy(), columns=["aggregate_signal"], index=index
        )

        df_signal = df_signal.reset_index(level=0)
        df_signal = df_signal.rename({"index": ""}, axis="columns")
        dates = [pd.Timestamp(d) for d in df_signal[""]]
        df_signal[""] = np.array(dates)

        plt.style.use("ggplot")

        fig, ax = plt.subplots()
        df_signal.plot.bar(
            x="", y="aggregate_signal", ax=ax, title=title, ylabel=y_label, legend=False
        )

        ticklabels = [""] * len(df_signal)
        skip = len(df_signal) // 12
        ticklabels[::skip] = df_signal[""].iloc[::skip].dt.strftime("%Y-%m-%d")
        ax.xaxis.set_major_formatter(mticker.FixedFormatter(ticklabels))

        fig.autofmt_xdate()

        def fmt(x, pos=0, max_i=len(ticklabels) - 1):
            i = int(x)
            i = 0 if i < 0 else max_i if i > max_i else i
            return dates[i]

        ax.fmt_xdata = fmt
        plt.show()

    def evaluate_pnls(
        self,
        pnl_cats: List[str],
        pnl_cids: List[str] = ["ALL"],
        start: Optional[str] = None,
        end: Optional[str] = None,
        label_dict: Optional[Dict[str, str]] = None,
    ):
        """
        Table of key PnL statistics.

        :param <List[str]> pnl_cats: list of PnL categories that should be plotted.
        :param <List[str]> pnl_cids: list of cross-sections to be plotted; default is
            'ALL' (global PnL).
            Note: one can only have multiple PnL categories or multiple cross-sections,
            not both.
        :param <str> start: earliest date in ISO format. Default is None and earliest
            date in df is used.
        :param <str> end: latest date in ISO format. Default is None and latest date
            in df is used.
        :param <dict[str, str]> label_dict: dictionary with keys as pnl_cats and values
            as new labels for the PnLs.

        :return <pd.DataFrame>: standardized DataFrame with key PnL performance
            statistics.
        """

        error_cids = "List of cross-sections expected."
        error_xcats = "List of PnL categories expected."
        assert isinstance(pnl_cids, list), error_cids
        assert isinstance(pnl_cats, list), error_xcats
        assert all([isinstance(elem, str) for elem in pnl_cids]), error_cids
        assert all([isinstance(elem, str) for elem in pnl_cats]), error_xcats

        if pnl_cats is None:
            # The field, self.pnl_names, is a data structure that stores the name of the
            # category assigned to PnL values. Each time make_pnl() method is called, the
            # computed DataFrame will have an associated category established by the
            # logical method: ('PNL_' + sig) if pnl_name is None else pnl_name. Each
            # category will be held in the data structure.
            pnl_cats = self.pnl_names
        else:
            if not set(pnl_cats) <= set(self.pnl_names):
                missing = [pnl for pnl in pnl_cats if pnl not in self.pnl_names]
                pnl_error = (
                    f"Received PnL categories have not been defined. The PnL "
                    f"category(s) which has not been defined is: {missing}. "
                    f"The produced PnL category(s) are {self.pnl_names}."
                )
                raise ValueError(pnl_error)

        assert (len(pnl_cats) == 1) | (len(pnl_cids) == 1)

        dfx = reduce_df(
            self.df, pnl_cats, pnl_cids, start, end, self.black, out_all=False
        )

        groups = "xcat" if len(pnl_cids) == 1 else "cid"
        stats = [
            "Return %",
            "St. Dev. %",
            "Sharpe Ratio",
            "Sortino Ratio",
            "Max 21-Day Draw %",
            "Max 6-Month Draw %",
            "Peak to Trough Draw %",
            "Top 5% Monthly PnL Share",
            "Traded Months",
        ]

        # If benchmark tickers have been passed into the Class and if the tickers are
        # present in self.dfd.
        benchmark_tickers = []

        if self.bm_bool and bool(self._bm_dict):
            benchmark_tickers = list(self._bm_dict.keys())
            for bm in benchmark_tickers:
                stats.insert(len(stats) - 1, f"{bm} correl")

        dfw = dfx.pivot(index="real_date", columns=groups, values="value")
        df = pd.DataFrame(columns=dfw.columns, index=stats)

        df.iloc[0, :] = dfw.mean(axis=0) * 261
        df.iloc[1, :] = dfw.std(axis=0) * np.sqrt(261)
        df.iloc[2, :] = df.iloc[0, :] / df.iloc[1, :]
        dsd = dfw.apply(lambda x: np.sqrt(np.sum(x[x < 0] ** 2) / len(x))) * np.sqrt(
            261
        )
        df.iloc[3, :] = df.iloc[0, :] / dsd
        df.iloc[4, :] = dfw.rolling(21).sum().min()
        df.iloc[5, :] = dfw.rolling(6 * 21).sum().min()

        cum_pnl = dfw.cumsum()
        high_watermark = cum_pnl.cummax()
        drawdown = high_watermark - cum_pnl

        df.iloc[6, :] = -drawdown.max()

        monthly_pnl = dfw.resample("M").sum()
        total_pnl = monthly_pnl.sum(axis=0)
        top_5_percent_cutoff = int(np.ceil(len(monthly_pnl) * 0.05))
        top_months = pd.DataFrame(columns=monthly_pnl.columns)
        for column in monthly_pnl.columns:
            top_months[column] = (
                monthly_pnl[column]
                .nlargest(top_5_percent_cutoff)
                .reset_index(drop=True)
            )

        df.iloc[7, :] = top_months.sum() / total_pnl

        if len(benchmark_tickers) > 0:
            bm_df = pd.concat(list(self._bm_dict.values()), axis=1)
            for i, bm in enumerate(benchmark_tickers):
                index = dfw.index.intersection(bm_df.index)
                correlation = dfw.loc[index].corrwith(
                    bm_df.loc[index].iloc[:, i], axis=0, method="pearson", drop=True
                )
                df.iloc[8 + i, :] = correlation

        df.iloc[8 + len(benchmark_tickers), :] = dfw.resample("M").sum().count()

        if label_dict is not None:
            if not isinstance(label_dict, dict):
                raise TypeError("label_dict must be a dictionary.")
            if not all([isinstance(k, str) for k in label_dict.keys()]):
                raise TypeError("Keys in label_dict must be strings.")
            if not all([isinstance(v, str) for v in label_dict.values()]):
                raise TypeError("Values in label_dict must be strings.")
            if len(label_dict) != len(df.columns):
                raise ValueError(
                    "label_dict must have the same number of keys as columns in the "
                    "DataFrame."
                )
            df.rename(index=label_dict, inplace=True)
            df = df[label_dict.values()]

        return df

    def print_pnl_names(self):
        """
        Print list of names of available PnLs in the class instance.
        """

        print(self.pnl_names)

    def pnl_df(self, pnl_names: List[str] = None, cs: bool = False):
        """
        Return dataframe with PnLs.

        :param <List[str]> pnl_names: list of names of PnLs to be returned.
            Default is 'ALL'.
        :param <bool> cs: inclusion of cross section PnLs. Default is False.

        :return <pd.DataFrame>: custom DataFrame with PnLs
        """
        selected_pnls = pnl_names if pnl_names is not None else self.pnl_names

        filter_1 = self.df["xcat"].isin(selected_pnls)
        filter_2 = self.df["cid"] == "ALL" if not cs else True

        return self.df[filter_1 & filter_2]

    def _winsorize(self, df: pd.DataFrame, thresh: float):
        if thresh is not None:
            df.clip(lower=-thresh, upper=thresh, inplace=True)


def create_results_dataframe(
    title: str,
    df: pd.DataFrame,
    ret: str,
    sigs: Union[str, List[str]],
    cids: Union[str, List[str]],
    sig_ops: Union[str, List[str]],
    sig_adds: Union[float, List[float]],
    neutrals: Union[str, List[str]],
    threshs: Union[float, List[float]],
    bm: str = None,
    sig_negs: Union[bool, List[bool]] = None,
    cosp: bool = False,
    start: str = None,
    end: str = None,
    blacklist: dict = None,
    freqs: Union[str, List[str]] = "M",
    agg_sigs: Union[str, List[str]] = "last",
    sigs_renamed: dict = None,
    fwin: int = 1,
    slip: int = 0,
):
    """
    Create a DataFrame with key performance metrics for the signals and PnLs.

    :param <str> title: title of the DataFrame.
    :param <pd.DataFrame> df: DataFrame with the data.
    :param <str> ret: name of the return signal.
    :param <Union[str, List[str] sigs: name of the comparative signal(s).
    :param <Union[str, List[str] cids: name of the cross-section(s).
    :param <Union[str, List[str] sig_ops: operation(s) to be applied to the signal(s).
    :param <Union[float, List[float] sig_adds: value(s) to be added to the signal(s).
    :param <Union[str, List[str] neutrals: neutralization method(s) to be applied.
    :param <Union[float, List[float] threshs: threshold(s) to be applied to the signal(s).
    :param <str> bm: name of the benchmark signal.
    :param <Union[bool, List[bool] sig_negs: whether the signal(s) should be negated.
    :param <bool> cosp: whether the signals should be cross-sectionally standardized.
    :param <str> start: start date of the analysis.
    :param <str> end: end date of the analysis.
    :param <dict> blacklist: dictionary with the blacklisted dates.
    :param <Union[str, List[str] freqs: frequency of the rebalancing.
    :param <Union[str, List[str] agg_sigs: aggregation method(s) for the signal(s).
    :param <dict> sigs_renamed: dictionary with the renamed signals.
    :param <int> fwin: frequency of the rolling window.
    :param <int> slip: slippage to be applied to the PnLs.

    :return <pd.DataFrame>: DataFrame with the performance metrics.
    """
    # Get the signals table and isolate relevant performance metrics

    def check_list_type(type, var):
        return isinstance(var, list) and all(isinstance(item, type) for item in var)

    if not isinstance(ret, str):
        raise TypeError("The return signal must be a string.")
    if not isinstance(sigs, str) and not check_list_type(str, sigs):
        raise TypeError("The signals must be a string or a list of strings.")
    if not isinstance(cids, str) and not check_list_type(str, cids):
        raise TypeError("The cids must be a string or a list of strings.")
    if not isinstance(sig_ops, str) and not check_list_type(str, sig_ops):
        raise TypeError("The signal operations must be a string or a list of strings.")
    if not isinstance(sig_adds, (float, int)) and not check_list_type(
        (float, int), sig_adds
    ):
        raise TypeError("The signal additions must be a float or a list of floats.")
    if not isinstance(neutrals, str) and not check_list_type(str, neutrals):
        raise TypeError("The neutralizations must be a string or a list of strings.")
    if not isinstance(threshs, (float, int)) and not check_list_type(
        (float, int), threshs
    ):
        raise TypeError("The thresholds must be a float or a list of floats.")

    srr = SignalReturnRelations(
        df=df,
        rets=ret,
        sigs=sigs,
        cids=cids,
        sig_neg=sig_negs,
        cosp=cosp,
        start=start,
        end=end,
        blacklist=blacklist,
        freqs=freqs,
        agg_sigs=agg_sigs,
        fwin=fwin,
        slip=slip,
    )
    sigs_df = (
        srr.multiple_relations_table()
        .astype("float")[["accuracy", "bal_accuracy", "pearson", "kendall"]]
        .round(3)
    )

    # Get the evaluated PnL statistics and isolate relevant performance metrics
    freq_dict = {
        "D": "daily",
        "W": "weekly",
        "M": "monthly",
    }
    pnl_freq = lambda x: freq_dict[x] if x in freq_dict.keys() else "monthly"

    if sig_negs is None:
        sig_negs = [False] * len(sigs)

    pnl = NaivePnL(
        df=df,
        ret=ret,
        sigs=sigs,
        cids=cids,
        bms=bm,
        start=start,
        end=end,
        blacklist=blacklist,
    )
    for idx, sig in enumerate(sigs):
        pnl.make_pnl(
            sig=sig,
            sig_op=(sig_ops if isinstance(sig_ops, str) else sig_ops[idx]),
            sig_add=(sig_adds if isinstance(sig_adds, (float, int)) else sig_adds[idx]),
            sig_neg=(sig_negs if isinstance(sig_negs, bool) else sig_negs[idx]),
            pnl_name=sig,
            rebal_freq=(
                pnl_freq(freqs) if isinstance(freqs, str) else pnl_freq(freqs[idx])
            ),
            rebal_slip=slip,
            vol_scale=10,
            neutral=(neutrals if isinstance(neutrals, str) else neutrals[idx]),
            thresh=(threshs if isinstance(threshs, (float, int)) else threshs[idx]),
        )

    if bm is not None:
        pnl_df = (
            pnl.evaluate_pnls(pnl_cats=sigs)
            .transpose()[["Sharpe Ratio", "Sortino Ratio", f"{bm} correl"]]
            .astype("float")
            .round(3)
        )
    else:
        pnl_df = (
            pnl.evaluate_pnls(pnl_cats=sigs)
            .transpose()[["Sharpe Ratio", "Sortino Ratio"]]
            .astype("float")
            .round(3)
        )

    # Sort pnl_df so that it matches the order of sigs
    pnl_df = pnl_df.reindex(sigs)

    sigs_df = sigs_df.reset_index(
        level=["Return", "Frequency", "Aggregation"], drop=True
    )

    if True in sig_negs:
        for sig in pnl_df.index:
            sig_neg = sig_negs[pnl_df.index.get_loc(sig)]
            if sig_neg:
                # Ensure each index is renamed only once
                pnl_df.rename(
                    index={sig: f"{sig}_NEG"},
                    inplace=True,
                )

    # Concatenate them and clean them
    res_df = pd.concat([sigs_df, pnl_df], axis=1)

    metric_map = {
        "Sharpe Ratio": "Sharpe",
        "Sortino Ratio": "Sortino",
        "accuracy": "Accuracy",
        "bal_accuracy": "Bal. Accuracy",
        "pearson": "Pearson",
        "kendall": "Kendall",
    }
    if bm is not None:
        metric_map[f"{bm} correl"] = "Market corr."

    res_df.rename(columns=metric_map, inplace=True)

    if sigs_renamed:
        res_df.rename(index=sigs_renamed, inplace=True)

    res_df = (
        res_df.style.format("{:.3f}")
        .set_caption(title)
        .set_table_styles(
            [
                {
                    "selector": "caption",
                    "props": [("text-align", "center"), ("font-weight", "bold")],
                }
            ]
        )
    )

    return res_df


@dataclass
class PnLParams:
    """
    Dataclass to store the parameters for the PnL creation.
    """

    signal: str
    sig_op: str
    sig_neg: bool
    sig_add: float
    pnl_name: str
    rebal_freq: str
    rebal_slip: int
    vol_scale: int
    neutral: str
    thresh: float


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "NZD", "USD", "EUR"]
    xcats = ["EQXR_NSA", "CRY", "GROWTH", "INFL", "DUXR"]

    cols_1 = ["earliest", "latest", "mean_add", "sd_mult"]
    df_cids = pd.DataFrame(index=cids, columns=cols_1)
    df_cids.loc["AUD", :] = ["2008-01-03", "2020-12-31", 0.5, 2]
    df_cids.loc["CAD", :] = ["2010-01-03", "2020-11-30", 0, 1]
    df_cids.loc["GBP", :] = ["2012-01-03", "2020-11-30", -0.2, 0.5]
    df_cids.loc["NZD"] = ["2002-01-03", "2020-09-30", -0.1, 2]
    df_cids.loc["USD"] = ["2015-01-03", "2020-12-31", 0.2, 2]
    df_cids.loc["EUR"] = ["2008-01-03", "2020-12-31", 0.1, 2]

    cols_2 = cols_1 + ["ar_coef", "back_coef"]

    df_xcats = pd.DataFrame(index=xcats, columns=cols_2)
    df_xcats.loc["EQXR_NSA"] = ["2000-01-03", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2000-01-01", "2020-10-30", 1, 2, 0.95, 1]
    df_xcats.loc["GROWTH"] = ["2010-01-03", "2020-10-30", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2001-01-01", "2020-10-30", 1, 2, 0.8, 0.5]
    df_xcats.loc["DUXR"] = ["2000-01-01", "2020-12-31", 0.1, 0.5, 0, 0.1]

    black = {"AUD": ["2006-01-01", "2015-12-31"], "GBP": ["2022-01-01", "2100-01-01"]}
    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # Instantiate a new instance to test the long-only functionality.
    # Benchmarks are used to calculate correlation against PnL series.
    pnl = NaivePnL(
        dfd,
        ret="EQXR_NSA",
        sigs=["CRY", "GROWTH", "INFL"],
        cids=cids,
        start="2000-01-01",
        blacklist=black,
        bms=["EUR_EQXR_NSA", "USD_EQXR_NSA"],
    )

    pnl.make_pnl(
        sig="GROWTH",
        sig_op="zn_score_pan",
        sig_neg=True,
        sig_add=0.5,
        rebal_freq="monthly",
        vol_scale=5,
        rebal_slip=1,
        min_obs=250,
        thresh=2,
    )

    pnl.make_pnl(
        sig="INFL",
        sig_op="zn_score_pan",
        sig_neg=True,
        sig_add=0.5,
        rebal_freq="monthly",
        vol_scale=5,
        rebal_slip=1,
        min_obs=250,
        thresh=2,
    )

    pnl.make_long_pnl(vol_scale=10, label="Long")

    df_eval = pnl.evaluate_pnls(
        pnl_cats=["PNL_GROWTH_NEG", "PNL_INFL_NEG"],
        start="2015-01-01",
        end="2020-12-31",
    )

    pnl.signal_heatmap(pnl_name="PNL_GROWTH_NEG", pnl_cids=cids, freq="m")

    print(df_eval)

    pnl.agg_signal_bars(
        pnl_name="PNL_GROWTH_NEG",
        freq="m",
        metric="direction",
        title=None,
    )
    pnl.plot_pnls(
        pnl_cats=["PNL_GROWTH_NEG", "Long"],
        title_fontsize=60,
        xlab="date",
        ylab="%",
    )
    pnl.plot_pnls(
        pnl_cats=["PNL_GROWTH_NEG", "Long"],
        facet=False,
        xcat_labels=["S_1", "S_2"],
        xlab="date",
        ylab="%",
    )
    pnl.plot_pnls(
        pnl_cats=["PNL_GROWTH_NEG", "Long"], facet=True, xcat_labels=["S_1", "S_2"]
    )
    pnl.plot_pnls(
        pnl_cats=["PNL_GROWTH_NEG", "Long"],
        facet=True,
    )

    pnl.plot_pnls(pnl_cats=["PNL_GROWTH_NEG"], pnl_cids=cids, xcat_labels=None)

    pnl.plot_pnls(
        pnl_cats=["PNL_GROWTH_NEG"], pnl_cids=cids, facet=True, xcat_labels=None
    )

    pnl.plot_pnls(
        pnl_cats=["PNL_GROWTH_NEG"],
        pnl_cids=cids,
        same_y=True,
        facet=True,
        xcat_labels=None,
        share_axis_labels=False,
        xlab="Date",
        ylab="PnL",
        y_label_adj=0.1,
    )

    results_eq_ols = create_results_dataframe(
        title="Performance metrics, PARITY vs OLS, equity",
        df=dfd,
        ret="EQXR_NSA",
        sigs=["GROWTH", "INFL", "CRY", "DUXR"],
        cids=cids,
        sig_ops="zn_score_pan",
        sig_adds=0,
        neutrals="zero",
        threshs=2,
        sig_negs=[True, False, False, True],
        bm="USD_EQXR_NSA",
        cosp=True,
        start="2004-01-01",
        freqs="M",
        agg_sigs="last",
        slip=1,
    )
    print(results_eq_ols.data)



import pandas as pd
from typing import List, Optional, Tuple
import seaborn as sns
import matplotlib.pyplot as plt

def view_table(
    df: pd.DataFrame,
    title: Optional[str] = None,
    title_fontsize: Optional[int] = 16,
    figsize: Optional[Tuple[float]] = (14, 4),
    min_color: float = -1,
    max_color: float = 1,
    xlabel: Optional[str] = None,
    ylabel: Optional[str] = None,
    xticklabels: Optional[List[str]] = None,
    yticklabels: Optional[List[str]] = None,
    annot: bool = True,
    fmt: str = ".2f",
) -> None:
    """
    Displays a DataFrame representing a table as a heatmap.

    :param <pd.Dataframe> table: table to be displayed.
    :param <str> title: string of chart title; defaults depend on type of range plot.
    :param <int> title_fontsize: font size of chart header. Default is 16.
    :param <Tuple[float]> figsize: Tuple (w, h) of width and height of plot.
    :param <float> min_color: minimum value of colorbar.
    :param <float> max_color: maximum value of colorbar.
    :param <str> xlabel: string of x-axis label.
    :param <str> ylabel: string of y-axis label.
    :param <List[str]> xticklabels: list of strings to label x-axis ticks.
    :param <List[str]> yticklabels: list of strings to label y-axis ticks.
    :param <bool> annot: whether to annotate heatmap with values.
    """

    if not isinstance(df, pd.DataFrame):
        raise TypeError("Table must be a DataFrame")

    if df.empty:
        raise ValueError("Table must not be empty")

    try:
        df = df.astype(float)
    except ValueError:
        raise ValueError("Table must be numeric")

    if xticklabels is None:
        xticklabels = df.columns.to_list()
    elif len(xticklabels) != len(df.columns):
        raise ValueError("Number of xticklabels must match number of columns")

    if yticklabels is None:
        yticklabels = df.index.to_list()
    elif len(yticklabels) != len(df.index):
        raise ValueError("Number of yticklabels must match number of rows")

    ax: plt.Axes
    fig, ax = plt.subplots(figsize=figsize)
    sns.set(style="ticks")
    sns.heatmap(
        df,
        cmap="vlag_r",
        vmin=min_color,
        vmax=max_color,
        square=False,
        linewidths=0.5,
        cbar_kws={"shrink": 0.5},
        annot=annot,
        fmt=fmt,
        xticklabels=xticklabels,
        yticklabels=yticklabels,
    )

    ax.set(xlabel=xlabel, ylabel=ylabel)
    ax.set_title(title, fontsize=title_fontsize)
    plt.tight_layout()

    plt.show()


if __name__ == "__main__":
    data = {
        "Col1": [1, 2, 3, 4],
        "Col2": [5, 6, 7, 8],
        "Col3": [9, 10, 11, 12],
        "Col4": [13, 14, 15, 16],
    }
    row_labels = ["Row1", "Row2", "Row3", "Row4"]

    df = pd.DataFrame(data, index=row_labels)

    view_table(df, title="Table")



# from .plots import FacetPlot, LinePlot, Plotter, Heatmap
from .facetplot import FacetPlot
from .lineplot import LinePlot
from .plotter import Plotter
from .heatmap import Heatmap
from .grades import view_grades
from .metrics import view_metrics
from .timelines import timelines
from .view_panel_dates import view_panel_dates
from .correlation import view_correlation
from .ranges import view_ranges
from .table import view_table
from .multiple_reg_scatter import multiple_reg_scatter
from .score_visualisers import ScoreVisualisers


__all__ = [
    "view_correlation",
    "FacetPlot",
    "view_grades",
    "Heatmap",
    "LinePlot",
    "view_metrics",
    "Plotter",
    "view_ranges",
    "view_table",
    "timelines",
    "view_panel_dates",
    "multiple_reg_scatter",
    "ScoreVisualisers",
]



"""
Functionality to visualize time series data as line charts.
"""
import numpy as np
import pandas as pd
from typing import List, Tuple, Optional

from macrosynergy.management.simulate import make_qdf

import macrosynergy.visuals as msv


def view_timelines(
    df: pd.DataFrame,
    xcats: Optional[List[str]] = None,
    cids: Optional[List[str]] = None,
    intersect: bool = False,
    val: str = "value",
    cumsum: bool = False,
    start: str = "2000-01-01",
    end: Optional[str] = None,
    ncol: int = 3,
    legend_ncol: int = 1,
    same_y: bool = True,
    all_xticks: bool = False,
    xcat_grid: bool = False,
    xcat_labels: Optional[List[str]] = None,
    single_chart: bool = False,
    label_adj: float = 0.05,
    title: Optional[str] = None,
    title_adj: float = 1.005,
    title_xadj: float = 0.5,
    title_fontsize: int = 18,
    cs_mean: bool = False,
    size: Tuple[float, float] = (12, 7),
    aspect: float = 1.618,
    height: float = 2.85,
    legend_fontsize: int = 12,
):
    """Displays a facet grid of time line charts of one or more categories.

    :param <pd.Dataframe> df: standardized DataFrame with the necessary columns:
        'cid', 'xcat', 'real_date' and at least one column with values of interest.
    :param <List[str]> xcats: extended categories to plot. Default is all in DataFrame.
    :param <List[str]> cids: cross sections to plot. Default is all in DataFrame.
        If this contains only one cross section a single line chart is created.
    :param <bool> intersect: if True only retains cids that are available for all xcats.
        Default is False.
    :param <str> val: name of column that contains the values of interest.
        Default is 'value'.
    :param <bool> cumsum: plot cumulative sum of the values over time. Default is False.
    :param <str> start: earliest date in ISO format. Default is earliest date available.
    :param <str> end: latest date in ISO format. Default is latest date available.
    :param <int> ncol: number of columns in facet grid. Default is 3.
    :param <int> legend_ncol: number of columns in legend. Default is 1.
    :param <bool> same_y: if True (default) all plots in facet grid share same y axis.
    :param <bool> all_xticks:  if True x-axis tick labels are added to all plots in grid.
        Default is False, i.e only the lowest row displays the labels.
    :param <bool> xcat_grid: if True, shows a facet grid of line charts for each xcat
        for a single cross section. Default is False, only one cross section is allowed
        with this option.
    :param <List[str]> xcat_labels: labels to be used for xcats. If not defined, the
        labels will be identical to extended categories.
    :param <bool> single_chart: if True, all lines are plotted in a single chart.
    :param <str> title: chart heading. Default is no title.
    :param <float> title_adj: parameter that sets top of figure to accommodate title.
        Default is 0.95.
    :param <float> title_xadj: parameter that sets x position of title. Default is 0.5.
    :param <int> title_fontsize: font size of title. Default is 16.
    :param <float> label_adj: parameter that sets bottom of figure to fit the label.
        Default is 0.05.
    :param <bool> cs_mean: if True this adds a line of cross-sectional averages to
        the line charts. This is only allowed for function calls with a single
        category. Default is False.
    :param <Tuple[float]> size: two-element tuple setting width/height of single cross
        section plot. Default is (12, 7). This is irrelevant for facet grid.
    :param <float> aspect: width-height ratio for plots in facet. Default is 1.7.
    :param <float> height: height of plots in facet. Default is 3.
    :param <int> legend_fontsize: font size of legend. Default is 12.

    """
    msv.timelines(
        df=df,
        xcats=xcats,
        cids=cids,
        intersect=intersect,
        val=val,
        cumsum=cumsum,
        start=start,
        end=end,
        ncol=ncol,
        same_y=same_y,
        all_xticks=all_xticks,
        xcat_grid=xcat_grid,
        xcat_labels=xcat_labels,
        single_chart=single_chart,
        title=title,
        legend_ncol=legend_ncol,
        cs_mean=cs_mean,
        title_fontsize=title_fontsize,
        legend_fontsize=legend_fontsize,
        # HC Params
        aspect=aspect,
        title_adj=title_adj,
        height=height,
        size=size,
        label_adj=label_adj,
        title_xadj=title_xadj,
    )


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "NZD"]
    xcats = ["XR", "CRY", "INFL", "FXXR"]
    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )
    df_cids.loc["AUD"] = ["2010-01-01", "2020-12-31", 0.2, 0.2]
    df_cids.loc["CAD"] = ["2011-01-01", "2020-11-30", 0, 1]
    df_cids.loc["GBP"] = ["2012-01-01", "2020-11-30", 0, 2]
    df_cids.loc["NZD"] = ["2012-01-01", "2020-09-30", -0.1, 3]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )

    df_xcats.loc["XR"] = ["2010-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["INFL"] = ["2015-01-01", "2020-12-31", 0.1, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2013-01-01", "2020-10-30", 1, 2, 0.95, 0.5]
    df_xcats.loc["FXXR"] = ["2013-01-01", "2020-10-30", 1, 2, 0.95, 0.5]

    dfd: pd.DataFrame = make_qdf(df_cids, df_xcats, back_ar=0.75)
    # sort by cid, xcat, and real_date
    dfd = dfd.sort_values(["cid", "xcat", "real_date"])
    ctr: int = -1
    for xcat in xcats[:2]:
        for cid in cids[:2]:
            ctr *= -1
            mask = (dfd["cid"] == cid) & (dfd["xcat"] == xcat)
            dfd.loc[mask, "value"] = (
                10
                * ctr
                * np.arange(dfd.loc[mask, "value"].shape[0])
                / (dfd.loc[mask, "value"].shape[0] - 1)
            )

    dfdx = dfd[~((dfd["cid"] == "AUD") & (dfd["xcat"] == "XR"))]

    view_timelines(
        dfd,
        xcats=["XR", "CRY"],
        cids=cids[0],
        size=(10, 5),
        title="AUD Return and Carry",
        aspect=3
    )

    view_timelines(
        dfd,
        xcats=["XR", "CRY", "INFL"],
        cids=cids[0],
        xcat_grid=True,
        title_adj=0.8,
        xcat_labels=["Return", "Carry", "Inflation"],
        title="AUD Return, Carry & Inflation",
        aspect=3
    )

    view_timelines(dfd, xcats=["CRY"], cids=cids, ncol=4, title="Carry", cs_mean=True)

    view_timelines(
        dfd, xcats=["XR"], cids=cids[:2], ncol=2, cumsum=True, same_y=False, aspect=2
    )

    dfd = dfd.set_index("real_date")
    view_timelines(
        dfd,
        xcats=["XR"],
        cids=cids[:2],
        ncol=2,
        cumsum=True,
        same_y=False,
        aspect=2,
        single_chart=True,
    )



"""
Contains mathematical utility functions used across the package.
"""

from typing import List

import pandas as pd
import numpy as np
import itertools
from macrosynergy.management.simulate import make_qdf


def expanding_mean_with_nan(
    dfw: pd.DataFrame, absolute: bool = False
) -> List[np.float64]:
    """
    Computes a rolling median of a vector of floats and returns the results. NaNs will be
    consumed.

    :param <QuantamentalDataFrame> dfw: "wide" dataframe with time index and 
        cross-sections as columns.
    :param <bool> absolute: if True, the rolling mean will be computed on the magnitude
        of each value. Default is False.

    :return <List[float] ret: a list containing the median values. The number of computed
        median values held inside the list will correspond to the number of timestamps
        the series is defined over.
    """

    if not isinstance(dfw, pd.DataFrame):
        raise TypeError("Method expects to receive a pd.DataFrame.")

    # cast the index to pd.Timestamp, if error raise TypeError
    try:
        dfw.index = pd.to_datetime(dfw.index)
    except:
        raise TypeError("The index of the DataFrame must be of type `<pd.Timestamp>`.")

    if not isinstance(absolute, bool):
        raise TypeError("The parameter `absolute` must be of type `<bool>`.")

    data: np.ndarray = dfw.to_numpy()

    no_rows: int = dfw.shape[0]
    no_columns: int = dfw.shape[1]
    no_elements: int = no_rows * no_columns

    one_dimension_arr: np.ndarray = data.reshape(no_elements)
    if absolute:
        one_dimension_arr = np.absolute(one_dimension_arr)

    rolling_summation: List[float] = [
        np.nansum(one_dimension_arr[0 : (no_columns * i)])
        for i in range(1, no_rows + 1)
    ]

    # Determine the number of active cross-sections per timestamp. Required for computing
    # the rolling mean.
    data_arr = data.astype(dtype=np.float32)
    # Sum across the rows.
    active_cross = np.sum(~np.isnan(data_arr), axis=1)
    rolling_active_cross = list(itertools.accumulate(active_cross))

    mean_calc = lambda m, d: m / d
    ret = list(map(mean_calc, rolling_summation, rolling_active_cross))

    return np.array(ret)


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP", "USD", "NZD"]
    xcats = ["XR", "CRY", "GROWTH", "INFL"]

    df_cids = pd.DataFrame(
        index=cids, columns=["earliest", "latest", "mean_add", "sd_mult"]
    )

    # Define the cross-sections over different timestamps such that the pivoted DataFrame
    # includes NaN values: more realistic testcase.
    df_cids.loc["AUD"] = ["2022-01-01", "2022-02-01", 0.5, 2]
    df_cids.loc["CAD"] = ["2022-01-10", "2022-02-01", 0.5, 2]
    df_cids.loc["GBP"] = ["2022-01-20", "2022-02-01", -0.2, 0.5]
    df_cids.loc["USD"] = ["2022-01-01", "2022-02-01", -0.2, 0.5]
    df_cids.loc["NZD"] = ["2022-01-05", "2022-02-01", -0.1, 2]

    df_xcats = pd.DataFrame(
        index=xcats,
        columns=["earliest", "latest", "mean_add", "sd_mult", "ar_coef", "back_coef"],
    )
    df_xcats.loc["XR"] = ["2010-01-01", "2022-02-01", 0, 1, 0, 0.3]
    df_xcats.loc["CRY"] = ["2011-01-01", "2022-02-01", 1, 2, 0.9, 0.5]
    df_xcats.loc["GROWTH"] = ["2012-01-01", "2022-02-01", 1, 2, 0.9, 1]
    df_xcats.loc["INFL"] = ["2013-01-01", "2022-02-01", 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd_xr = dfd[dfd["xcat"] == "XR"]

    dfw = dfd_xr.pivot(index="real_date", columns="cid", values="value")
    no_rows = dfw.shape[0]

    ret_mean = expanding_mean_with_nan(dfw)



"""
A subclass inheriting from `macrosynergy.visuals.plotter.Plotter`,
designed to plot time series data as a heatmap.
"""

import pandas as pd
from typing import List, Tuple, Optional
import matplotlib.pyplot as plt
import numpy as np
from seaborn.utils import relative_luminance
import matplotlib as mpl
from typing import Union
from macrosynergy.management.utils import downsample_df_on_real_date

from macrosynergy.visuals.plotter import Plotter
from numbers import Number
from macrosynergy.management.simulate import make_test_df


class Heatmap(Plotter):
    """
    Class for plotting time series data as a heatmap.
    Inherits from `macrosynergy.visuals.plotter.Plotter`.

    Parameters
    :param <pd.DataFrame> df: A DataFrame with the following columns:
        'cid', 'xcat', 'real_date', and at least one metric from -
        'value', 'grading', 'eop_lag', or 'mop_lag'.
    :param <List[str]> cids: A list of cids to select from the DataFrame.
        If None, all cids are selected.
    :param <List[str]> xcats: A list of xcats to select from the DataFrame.
        If None, all xcats are selected.
    :param <List[str]> metrics: A list of metrics to select from the DataFrame.
        If None, all metrics are selected.
    :param <str> start: ISO-8601 formatted date. Select data from
        this date onwards. If None, all dates are selected.
    :param <str> end: ISO-8601 formatted date. Select data up to
        and including this date. If None, all dates are selected.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        cids: Optional[List[str]] = None,
        xcats: Optional[List[str]] = None,
        metrics: Optional[List[str]] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            df=df,
            cids=cids,
            xcats=xcats,
            metrics=metrics,
            start=start,
            end=end,
            *args,
            **kwargs,
        )

    def _plot(
        self,
        df: pd.DataFrame,
        figsize: Tuple[Number, Number] = (12, 8),
        x_axis_label: Optional[str] = None,
        y_axis_label: Optional[str] = None,
        axis_fontsize: int = 14,
        title: Optional[str] = None,
        title_fontsize: int = 22,
        title_xadjust: Number = 0.5,
        title_yadjust: Number = 1.0,
        vmin: Optional[Number] = None,
        vmax: Optional[Number] = None,
        show: bool = True,
        save_to_file: Optional[str] = None,
        dpi: int = 300,
        return_figure: bool = False,
        on_axis: Optional[plt.Axes] = None,
        max_xticks: int = 50,
        cmap: Optional[Union[str, mpl.colors.Colormap]] = None,
        rotate_xticks: Optional[Number] = 0,
        rotate_yticks: Optional[Number] = 0,
        show_tick_lines: Optional[bool] = True,
        show_colorbar: Optional[bool] = True,
        show_annotations: Optional[bool] = False,
        show_boundaries: Optional[bool] = False,
        annotation_fontsize: int = 14,
        tick_fontsize: int = 13,
        *args,
        **kwargs,
    ) -> Optional[plt.Figure]:
        """
        Plots a DataFrame as a heatmap with the columns along the x-axis and
        rows along the y-axis.

        Parameters
        :param <Tuple> figsize: tuple specifying the size of the figure. Default is
            (12, 8).
        :param <str> x_axis_label: label for x-axis.
        :param <str> y_axis_label: label for y-axis.
        :param <int> axis_fontsize: the font size for the axis labels.
        :param <str> title: the figure's title.
        :param <int> title_fontsize: the font size for the title.
        :param <float> title_xadjust: sets the x position of the title text.
        :param <float> title_yadjust: sets the y position of the title text.
        :param <float> vmin: optional minimum value for heatmap scale.
        :param <float> vmax: optional maximum value for heatmap scale.
        :param <bool> show: if True, the image is displayed.
        :param <str> save_to_file: the path at which to save the heatmap as an image.
            If not specified, the plot will not be saved.
        :param <int> dpi: the resolution in dots per inch used if saving the figure.
        :param <bool> return_figure: if True, the function will return the figure.
        :param <plt.Axes> on_axis: optional `plt.Axes` object to be used instead of
            creating a new one.
        :param <int> max_xticks: the maximum number of ticks to be displayed
            along the x axis. Default is 50.
        :param <mpl.colors.Colormap> cmap: string or matplotlib Colormap object
            specifying the colormap of the plot.
        :param <int> rotate_xticks: number of degrees to rotate the tick labels on
            the x-axis. Default is zero.
        :param <int> rotate_yticks: number of degrees to rotate the tick labels on
            the y-axis. Default is zero.
        :param <bool> show_tick_lines: if True, lines are shown for ticks.
            Default is True.
        :param <bool> show_colorbar: if True, the colorbar is shown. Default is True.
        :param <bool> show_annotations: if True, annotations display the value of
            each cell. Default is False.
        :param <bool> show_boundaries: if True, cells are divided by a grid.
            Default is False.
        :param <int> annotation_fontsize: sets the font size of the annotations.
        :param <int> tick_fontsize: sets the font size of tick labels.
        """
        if on_axis:
            fig: plt.Figure = on_axis.get_figure()
            ax: plt.Axes = on_axis
        else:
            fig: plt.Figure
            ax: plt.Axes
            fig, ax = plt.subplots(figsize=figsize, layout="constrained")

        data = df.to_numpy()

        im = ax.imshow(
            data,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            aspect="auto",
            **kwargs,
        )

        xtick_labels = df.columns.to_list()
        ytick_labels = df.index.to_list()

        ax.set_xticks(np.arange(len(xtick_labels)), labels=xtick_labels)
        ax.set_yticks(np.arange(len(ytick_labels)), labels=ytick_labels)

        ax.set_xticklabels(
            xtick_labels,
            rotation=rotate_xticks,
            ha="center",
            minor=False,
        )
        ax.set_yticklabels(
            ytick_labels,
            rotation=rotate_yticks,
            ha="right",
            minor=False,
            rotation_mode="anchor",
        )

        ax.tick_params(axis="both", which="major", labelsize=tick_fontsize)

        if show_tick_lines:
            ax.tick_params(which="major", length=4, width=1, direction="out")

        plt.grid(False)

        if show_boundaries:
            ax.spines[:].set_visible(False)
            ax.set_xticks(np.arange(len(xtick_labels) + 1) - 0.5, minor=True)
            ax.set_yticks(np.arange(len(ytick_labels) + 1) - 0.5, minor=True)
            ax.grid(which="minor", color="w", linestyle="-", linewidth=3)
            ax.tick_params(which="minor", bottom=False, left=False)
        else:
            # Limits the number of ticks shown on the x-axis.
            ax.xaxis.set_major_locator(plt.MaxNLocator(nbins=max_xticks - 1))

        ax.set_title(
            title,
            fontsize=title_fontsize,
            x=title_xadjust,
            y=title_yadjust,
        )

        if x_axis_label:
            ax.set_xlabel(x_axis_label, fontsize=axis_fontsize)

        if y_axis_label:
            ax.set_ylabel(y_axis_label, fontsize=axis_fontsize)

        if show_colorbar:
            ax.figure.colorbar(im, ax=ax)

        if show_annotations:
            data = np.around(data, decimals=1)
            for i in range(data.shape[0]):
                for j in range(data.shape[1]):
                    color = im.cmap(im.norm(im.get_array()))[i, j]
                    lum = relative_luminance(color)
                    text_color = ".15" if lum > 0.408 else "w"
                    text_kwargs = dict(
                        color=text_color,
                        ha="center",
                        va="center",
                        size=annotation_fontsize,
                    )
                    if not np.isnan(data[i, j]):
                        ax.text(j, i, data[i, j], **text_kwargs)

        ax.tick_params(axis="y", which="major", pad=8)

        if save_to_file:
            plt.savefig(
                save_to_file,
                dpi=dpi,
                bbox_inches="tight",
            )

        if show:
            plt.show()
            return

        if return_figure:
            return fig

    def plot_metric(
        self,
        x_axis_column,
        y_axis_column,
        metric,
        xcats=None,
        cids=None,
        start=None,
        end=None,
        freq=None,
        agg="mean",
        figsize: Optional[Tuple[Number, Number]] = (12, 8),
        x_axis_label: Optional[str] = None,
        y_axis_label: Optional[str] = None,
        axis_fontsize: int = 14,
        title: Optional[str] = None,
        title_fontsize: int = 22,
        title_xadjust: Number = 0.5,
        title_yadjust: Number = 1.0,
        vmin: Optional[Number] = None,
        vmax: Optional[Number] = None,
        show: bool = True,
        save_to_file: Optional[str] = None,
        dpi: int = 300,
        return_figure: bool = False,
        on_axis: Optional[plt.Axes] = None,
        max_xticks: int = 50,
        cmap: Optional[Union[str, mpl.colors.Colormap]] = None,
        rotate_xticks: Optional[Number] = 0,
        rotate_yticks: Optional[Number] = 0,
        show_tick_lines: Optional[bool] = True,
        show_colorbar: Optional[bool] = True,
        show_annotations: Optional[bool] = False,
        show_boundaries: Optional[bool] = False,
        annotation_fontsize: int = 14,
        tick_fontsize: int = 13,
        *args,
        **kwargs,
    ):
        df = self.df.copy()
        if not xcats:
            xcats = self.xcats
        if not cids:
            cids = self.cids
        if not start:
            start = self.start
        if not end:
            end = self.end

        # Validation checks not covered by Plotter.
        if metric not in ["value", "eop_lag", "mop_lag", "grading"]:
            raise ValueError(
                "`metric` must be either 'eop_lag', 'mop_lag', 'grading', or 'value'"
            )

        if not isinstance(agg, str):
            raise TypeError("`agg` must be a string")
        else:
            agg: str = agg.lower()
            if agg not in ["mean", "median", "min", "max", "first", "last"]:
                raise ValueError(
                    "`agg` must be one of 'mean', 'median', 'min', 'max', 'first' or "
                    "'last'"
                )

        df = df[["xcat", "cid", "real_date", metric]]

        if freq:
            df: pd.DataFrame = downsample_df_on_real_date(
                df=df, groupby_columns=["cid", "xcat"], freq=freq, agg=agg
            )

        if "real_date" not in [x_axis_column, y_axis_column]:
            df = df.groupby(["xcat", "cid"]).mean().reset_index()

        vmax: float = max(1, df[metric].max())
        vmin: float = min(0, df[metric].min())

        df["real_date"] = df["real_date"].dt.strftime("%Y-%m-%d")

        df = df.pivot_table(index=y_axis_column, columns=x_axis_column, values=metric)

        if figsize is None:
            figsize = (
                max(df.shape[0] / 2, 15),
                max(1, df.shape[1] / 2),
            )
        elif isinstance(figsize, list):
            figsize = tuple(figsize)

        self._plot(
            df=df,
            figsize=figsize,
            x_axis_label=x_axis_label,
            y_axis_label=y_axis_label,
            axis_fontsize=axis_fontsize,
            title=title,
            title_fontsize=title_fontsize,
            title_xadjust=title_xadjust,
            title_yadjust=title_yadjust,
            vmin=vmin,
            vmax=vmax,
            show=show,
            save_to_file=save_to_file,
            dpi=dpi,
            return_figure=return_figure,
            on_axis=on_axis,
            max_xticks=max_xticks,
            cmap=cmap,
            rotate_xticks=rotate_xticks,
            rotate_yticks=rotate_yticks,
            show_tick_lines=show_tick_lines,
            show_colorbar=show_colorbar,
            show_annotations=show_annotations,
            show_boundaries=show_boundaries,
            annotation_fontsize=annotation_fontsize,
            tick_fontsize=tick_fontsize,
        )


if __name__ == "__main__":
    test_cids: List[str] = [
        "USD",
    ]  # "EUR", "GBP"]
    test_xcats: List[str] = ["FX", "IR"]
    dfE: pd.DataFrame = make_test_df(
        cids=test_cids, xcats=test_xcats, style="sharp-hill"
    )

    dfM: pd.DataFrame = make_test_df(
        cids=test_cids, xcats=test_xcats, style="four-bit-sine"
    )

    dfG: pd.DataFrame = make_test_df(cids=test_cids, xcats=test_xcats, style="sine")

    dfE.rename(columns={"value": "eop_lag"}, inplace=True)
    dfM.rename(columns={"value": "mop_lag"}, inplace=True)
    dfG.rename(columns={"value": "grading"}, inplace=True)
    mergeon = ["cid", "xcat", "real_date"]
    dfx: pd.DataFrame = pd.merge(pd.merge(dfE, dfM, on=mergeon), dfG, on=mergeon)

    heatmap = Heatmap(df=dfx, xcats=["FX"])

    heatmap.df["real_date"] = heatmap.df["real_date"].dt.strftime("%Y-%m-%d")
    heatmap.df = heatmap.df.pivot_table(
        index="cid", columns="real_date", values="grading"
    )

    heatmap._plot(heatmap.df, title="abc", rotate_xticks=90)



"""
A subclass inheriting from `macrosynergy.visuals.plotter.Plotter`,
designed to render a facet plot containing any number of subplots.
Given that the class allows returning a `matplotlib.pyplot.Figure`,
one can easily add any number of subplots, even the FacetPlot itself:
effectively allowing for a recursive facet plot.
"""

from typing import Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.gridspec import GridSpec

from macrosynergy.visuals.plotter import Plotter
from numbers import Number


def _get_square_grid(
    num_plots: int,
) -> Tuple[int, int]:
    """
    Given the number of plots, return a tuple of grid dimensions
    that is closest to a square grid.

    Parameters
    :param <int> num_plots: Number of plots.
    :return <Tuple[int, int]>: Tuple of grid dimensions.
    """
    sqrt_num_plots: float = np.ceil(np.sqrt(num_plots))
    grid_dim: Tuple[int, int] = (int(sqrt_num_plots), int(sqrt_num_plots))
    gd_copy: Tuple[int, int] = grid_dim
    # the number of plots is less than grid_dim[0] * grid_dim[1],
    # so iteratively try and reduce the row and column dimensions until sweet spot is
    # found.
    while 0 != 1:
        if gd_copy[0] < gd_copy[1]:
            gd_copy: Tuple[int, int] = (gd_copy[0], gd_copy[1] - 1)
        else:
            gd_copy: Tuple[int, int] = (gd_copy[0] - 1, gd_copy[1])

        # if gd_copy[0] * gd_copy[1] is more than num_plots, copy to grid_dim. if smaller,
        # break.
        if gd_copy[0] * gd_copy[1] >= num_plots:
            grid_dim: Tuple[int, int] = gd_copy
        else:
            break

    return grid_dim


class FacetPlot(Plotter):
    """
    Class for rendering a facet plot containing any number of subplots.
    Inherits from `macrosynergy.visuals.plotter.Plotter`.

    Parameters
    :param <pd.DataFrame> df: A DataFrame with the following columns:
        'cid', 'xcat', 'real_date', and at least one metric from -
        'value', 'grading', 'eop_lag', or 'mop_lag'.
    :param <List[str]> cids: A list of cids to select from the DataFrame.
        If None, all cids are selected.
    :param <List[str]> xcats: A list of xcats to select from the DataFrame.
        If None, all xcats are selected.
    :param <List[str]> metrics: A list of metrics to select from the DataFrame.
        If None, all metrics are selected.
    :param <bool> intersect: if True only retains cids that are available for
        all xcats. Default is False.
    :param <List[str]> tickers: A list of tickers to select from the DataFrame.
        If None, all tickers are selected.
    :param <dict> blacklist: cross-sections with date ranges that should be excluded from
        the data frame. If one cross-section has several blacklist periods append numbers
        to the cross-section code.
    :param <str> start: ISO-8601 formatted date string. Select data from
        this date onwards. If None, all dates are selected.
    :param <str> end: ISO-8601 formatted date string. Select data up to
        and including this date. If None, all dates are selected.

    """

    def __init__(
        self,
        df: Optional[pd.DataFrame] = None,
        cids: Optional[List[str]] = None,
        xcats: Optional[List[str]] = None,
        metrics: Optional[List[str]] = None,
        intersect: bool = False,
        tickers: Optional[List[str]] = None,
        blacklist: Optional[Dict[str, List[str]]] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            df=df,
            cids=cids,
            xcats=xcats,
            metrics=metrics,
            intersect=intersect,
            tickers=tickers,
            blacklist=blacklist,
            start=start,
            end=end,
            *args,
            **kwargs,
        )

    def _get_grid_dim(
        self,
        tickers: List[str],
        ncols: Optional[int] = None,
        attempt_square: bool = False,
        cid_grid: bool = False,
        xcat_grid: bool = False,
        cid_xcat_grid: bool = False,
        _cids: Optional[List[str]] = None,
        _xcats: Optional[List[str]] = None,
    ) -> Tuple[int, int]:
        """
        Returns a tuple of grid dimensions that matches the plot settings.
        """
        # only one of cid_grid, xcat_grid, cid_xcat_grid or attempt_square can be True
        if sum([cid_grid, xcat_grid, cid_xcat_grid]) > 1:
            raise ValueError(
                "Only one of `cid_grid`, `xcat_grid`, or "
                "`cid_xcat_grid` can be True."
            )

        if _cids is None:
            _cids: List[str] = self.cids
        if _xcats is None:
            _xcats: List[str] = self.xcats

        if attempt_square:
            target_var: str = tickers
            if cid_grid:
                target_var: List[str] = _cids
            elif xcat_grid:
                target_var: List[str] = _xcats
            return _get_square_grid(num_plots=len(target_var))

        if cid_grid or xcat_grid:
            tks: List[str] = _cids if cid_grid else _xcats
            return self._get_grid_dim(
                tickers=tks,
                ncols=ncols,
                attempt_square=attempt_square,
                _cids=_cids,
                _xcats=_xcats,
            )

        if cid_xcat_grid:
            return (len(_xcats), len(_cids))

        if ncols is not None:
            return (
                ncols,
                int(np.ceil(len(tickers) / ncols)),
            )

        raise ValueError("Unable to infer grid dimensions.")

    # def scatterplot(
    #     self,
    #     # plot arguments
    #     cids: Optional[List[str]] = None,
    #     xcats: Optional[List[str]] = None,
    #     metric: Optional[str] = None,
    #     # fig arguments
    #     figsize: Tuple[Number, Number] = (16.0, 9.0),
    #     ncols: int = 3,
    #     attempt_square: bool = False,
    #     # xcats_mean: bool = False,
    #     # title arguments
    #     title: Optional[str] = None,
    #     title_fontsize: int = 20,
    #     title_xadjust: Optional[Number] = None,
    #     title_yadjust: Optional[Number] = None,
    #     # subplot axis arguments
    #     ax_grid: bool = True,
    #     ax_hline: bool = False,
    #     ax_hline_val: Number = 0.0,
    #     ax_vline: bool = False,
    #     ax_vline_val: Number = 0.0,
    #     x_axis_label: Optional[str] = None,
    #     y_axis_label: Optional[str] = None,
    #     axis_fontsize: int = 12,
    #     # subplot arguments
    #     facet_size: Optional[Tuple[Number, Number]] = None,
    #     facet_titles: Optional[List[str]] = None,
    #     facet_title_fontsize: int = 12,
    #     facet_title_xadjust: Number = 0.5,
    #     facet_title_yadjust: Number = 1.0,
    #     facet_xlabel: Optional[str] = None,
    #     facet_ylabel: Optional[str] = None,
    #     facet_label_fontsize: int = 12,
    #     # legend arguments
    #     legend: bool = True,
    #     legend_labels: Optional[List[str]] = None,
    #     legend_loc: str = "upper center",
    #     legend_ncol: int = 1,
    #     legend_bbox_to_anchor: Optional[Tuple[Number, Number]] = None,  # (1.0, 0.5),
    #     legend_frame: bool = True,
    #     # return args
    #     show: bool = True,
    #     save_to_file: Optional[str] = None,
    #     dpi: int = 300,
    #     return_figure: bool = False,
    #     plot_func_args: Dict[str, Any] = {},
    #     *args,
    #     **kwargs,
    # ):
    #     """
    #     **NOT IMPLEMENTED YET**

    #     Showing a FacetPlot composed of scatter plots from the data available in the
    #     `FacetPlot` object after initialization.
    #     """
    #     raise NotImplementedError("Scatterplot not implemented yet.")
    #     if metric is None:
    #         metric: str = self.metrics[0]

    #     _cids: List[str] = self.cids if cids is None else cids
    #     _xcats: List[str] = self.xcats if xcats is None else xcats

    #     ...

    def lineplot(
        self,
        # plot arguments
        cids: Optional[List[str]] = None,
        xcats: Optional[List[str]] = None,
        metric: Optional[str] = None,
        ncols: int = 3,
        attempt_square: bool = False,
        cid_grid: bool = False,
        xcat_grid: bool = False,
        cid_xcat_grid: bool = False,
        grid_dim: Optional[Tuple[int, int]] = None,
        compare_series: Optional[str] = None,
        share_y: bool = False,
        share_x: bool = False,
        interpolate: bool = False,
        # xcats_mean: bool = False,
        # title arguments
        figsize: Tuple[Number, Number] = (16.0, 9.0),
        title: Optional[str] = None,
        title_fontsize: int = 22,
        title_xadjust: Optional[Number] = None,
        title_yadjust: Optional[Number] = None,
        # subplot axis arguments
        ax_grid: bool = False,
        ax_hline: Optional[Number] = 0.0,
        ax_vline: Optional[str] = None,
        x_axis_label: Optional[str] = None,
        y_axis_label: Optional[str] = None,
        axis_fontsize: int = 12,
        # subplot arguments
        facet_size: Optional[Tuple[Number, Number]] = None,
        facet_titles: Optional[List[str]] = None,
        facet_title_fontsize: int = 14,
        facet_title_xadjust: Number = 0.5,
        facet_title_yadjust: Number = 1.0,
        facet_xlabel: Optional[str] = None,
        facet_ylabel: Optional[str] = None,
        # legend arguments
        legend: bool = True,
        legend_labels: Optional[List[str]] = None,
        legend_loc: Optional[str] = "lower center",
        legend_fontsize: int = 12,
        legend_ncol: int = 1,
        legend_bbox_to_anchor: Optional[Tuple[Number, Number]] = None,  # (1.0, 0.5),
        legend_frame: bool = True,
        # return args
        show: bool = True,
        save_to_file: Optional[str] = None,
        dpi: int = 300,
        return_figure: bool = False,
        *args,
        **kwargs,
    ):
        """
        Showing a FacetPlot composed of linear plots from the data available in the
        `FacetPlot` object after initialization.
        Passing any of the arguments used to initialize the `FacetPlot` object will cause
        the object to be re-initialized with the new arguments, and the plot will be
        rendered from the new object state.

        Parameters
        :param <int> ncols: number of columns in the grid. Default is 3.
        :param <bool> attempt_square: attempt to make the facet grid square. Ignores
            `ncols` when `True`. Default is `False`.
        :param <bool> cid_grid: Create the facet grid such that each facet is a `cid`.
        :param <bool> xcat_grid: Create the facet grid such that each facet is an `xcat`.
        :param <bool> cid_xcat_grid: Create the facet grid such that each facet is an
            individual ticker. Each "row" contains plots for the same `cid`, and each
            "column" would contain plots for the same `xcat`. Therefore, this mode does
            not respect the `ncols` or `attempt_square` arguments.
            NB: `facet_titles` and `legend` are overridden in this mode.
        :param <bool> cids_mean: Used with `cid_grid` with a single `xcat`. If `True`,
            the mean of all `cids` for that `xcat` will be plotted on all charts. If
            `False`, only the specified `cids` will be plotted. Default is `False`.
        :param <str> compare_series: Used with `cid_grid` with a single `xcat`. If
            specified, the series specified will be plotted in each facet, as a red dashed
            line. This is useful for comparing a single series, such as a
            benchmark/average. Ensure that the comparison series is in the dataframe,
            and not filtered out when initializing the `FacetPlot` object. Default is
            `None`. NB: `compare_series` can only be used when the series is not removed
            by `reduce_df()` in the object initialization.
        :param <bool> share_y: whether to share the y-axis across all plots. Default is
            `True`.
        :param <bool> share_x: whether to share the x-axis across all plots. Default is
            `True`.
        :param <bool> interpolate: if `True`, gaps in the time series will be interpolated.
            Default is `False`.
        :param <Tuple[Number, Number]> figsize: a tuple of floats specifying the width
            and height of the figure. Default is `(16.0, 9.0)`.
        :param <str> title: the title of the plot. Default is `None`.
        :param <int> title_fontsize: the font size of the title. Default is `20`.
        :param <float> title_xadjust: the x-adjustment of the title. Default is `None`.
        :param <float> title_yadjust: the y-adjustment of the title. Default is `None`.
        :param <bool> ax_grid: whether to show the grid on the axes, applied to all plots.
            Default is `True`.
        :param <Number> ax_hline: the value of the horizontal line on the axes, applied
            to all plots. Default is `None`, meaning no horizontal line will be shown.
        :param <str> ax_vline: the value of the vertical line on the axes, applied
            to all plots. The value must be a ISO-8601 formatted date-string.
            Default is `None`, meaning no vertical line will be shown.
        :param <str> x_axis_label: the label for the x-axis. Default is `None`.
        :param <str> y_axis_label: the label for the y-axis. Default is `None`.
        :param <int> axis_fontsize: the font size of the axis labels. Default is `12`.
        :param <Tuple[Number, Number]> facet_size: a tuple of floats specifying the
            width and height of each facet. Default is `None`, meaning the facet size will
            be inferred from the `figsize` argument. If specified, the `figsize` argument
            will be ignored and the figure size will be inferred from the dimensions of
            the facet grid and the facet size.
        :param <List[str]> facet_titles: a list of strings specifying the titles of each
            facet. Default is `None`, meaning all facets will have the full `ticker`,
            `cid`, or `xcat` as the title. If no `facet_titles` are required,
            pass an empty list - `facet_titles=[]`.
        :param <int> facet_title_fontsize: the font size of the facet titles. Default is
            `12`.
        :param <float> facet_title_xadjust: the x-adjustment of the facet titles. Default
            is `None`.
        :param <float> facet_title_yadjust: the y-adjustment of the facet titles. Default
            is `None`.
        :param <bool> facet_xlabel: The label to be used as the axis-label/title for the
            x-axis of each facet. Default is `None`, meaning no label will be shown.
        :param <bool> facet_ylabel: The label to be used as the axis-label/title for the
            y-axis of each facet. Default is `None`, meaning no label will be shown.
        :param <int> facet_label_fontsize: the font size of the facet labels. Default is
            `12`.
        :param <bool> legend: Show the legend. Default is `True`. When using
            `cid_xcat_grid`, the legend will not be shown as it is redundant.
        :param <list> legend_labels: Labels for the legend. Default is `None`,
            meaning a list identifying the various `cids`/`xcats` will be used.
        :param <str> legend_loc: Location of the legend. Default is `center left`.
            See https://matplotlib.org/stable/api/_as_gen/matplotlib.pyplot.legend.html
            for more information.
        :param <int> legend_fontsize: Font size of the legend. Default is `12`.
        :param <int> legend_ncol: Number of columns in the legend. Default is `1`.
        :param <tuple> legend_bbox_to_anchor: Bounding box for the legend. Default is
            `(1.0, 0.5)`.
        :param <bool> legend_frame: Show the legend frame. Default is `True`.
        :param <bool> show: Show the plot. Default is `True`.
        :param <str> save_to_file: Save the plot to a file. Default is `None`.
        :param <int> dpi: DPI of the saved image. Default is `300`.
        :param <bool> return_figure: Return the figure object. Default is `False`.
        """
        comp_series_flag: bool = False

        if compare_series:
            if compare_series not in set(
                (kwargs["df"] if "df" in kwargs else self.df)[["cid", "xcat"]]
                .drop_duplicates()
                .apply(lambda x: "_".join(x), axis=1)
                .tolist()
            ):
                comp_series_flag: bool = True

        if metric is None:
            metric: str = self.metrics[0]

        _cids: List[str] = self.cids if cids is None else cids
        _xcats: List[str] = self.xcats if xcats is None else xcats

        tickers_to_plot: List[str] = (
            (self.df["cid"] + "_" + self.df["xcat"]).unique().tolist()
        )

        _tk: List[str] = tickers_to_plot.copy()
        if compare_series:
            _tk.remove(compare_series)
        grid_dim: Tuple[int, int] = self._get_grid_dim(
            tickers=_tk,
            ncols=ncols,
            attempt_square=attempt_square,
            cid_grid=cid_grid,
            xcat_grid=xcat_grid,
            cid_xcat_grid=cid_xcat_grid,
            _cids=cids,
            _xcats=xcats,
        )

        # if the facet size is not none, re-calc the figsize
        if facet_size is not None:
            figsize: Tuple[float, float] = (
                grid_dim[0] * facet_size[0] * 1.5,
                grid_dim[1] * facet_size[1] * 1.5,
            )
            # mul by 1.5 to account for the space taken up annotations, etc.
            # fig.tight_layout() cleans this up

        # form a dictionary of what goes on each plot
        # each key is an int - identifying the plot (L-R, T-B)
        # each value is a dict with keys "xs", "ys"

        if compare_series is not None:
            # check that the compare series exists in the dataframe
            if compare_series not in tickers_to_plot:
                raise ValueError(
                    f"Compare series {compare_series} not found in dataframe."
                )
            else:
                tickers_to_plot.remove(compare_series)

        plot_dict: Dict[str, Dict[str, Union[str, List[str]]]] = {}

        colormap = plt.cm.tab10
        legend_color_map: Optional[Dict[str, str]] = None

        if facet_titles is None:
            if cid_grid:
                facet_titles: List[str] = _cids
            elif xcat_grid:
                facet_titles: List[str] = _xcats
            elif cid_xcat_grid:
                # cid_xcat_grid facets only make sense if they have cid_xcat as the title
                legend: bool = False
            else:
                facet_titles: List[str] = tickers_to_plot

        if not any([cid_grid, xcat_grid, cid_xcat_grid]):
            # each ticker gets its own plot
            for i, ticker in enumerate(tickers_to_plot):
                tks: List[str] = [ticker]
                if compare_series is not None:
                    tks.append(compare_series)
                plot_dict[i] = {
                    "X": "real_date",
                    "Y": tks,
                }
                # plot_dict[i] --> Dict[str, Union[str, List[str]]]

        if cid_grid or xcat_grid:
            # flipper handles resolution between cid_grid and xcat_grid for binary
            # variables
            flipper: bool = 1 if cid_grid else -1
            if facet_titles is None:
                # needs to be "flipped" twice, as facet_titles need to be complementary
                # to the legend labels
                facet_titles: List[str] = [_cids, _xcats][::flipper][0]
            if legend_labels is None:
                legend_labels: List[str] = [_xcats, _cids][::flipper][0]
            elif len(legend_labels) != (
                len([_xcats, _cids][::flipper][0]) + bool(compare_series)
            ):
                raise ValueError(
                    "The number of legend labels does not match the lines to plot."
                )

            if legend_color_map is None:
                legend_color_map: Dict[str, str] = {
                    x: colormap(i) for i, x in enumerate([_xcats, _cids][::flipper][0])
                }

            for i, fvar in enumerate([_cids, _xcats][::flipper][0]):
                tks: List[str] = [
                    "_".join([fvar, x][::flipper])
                    for x in ([_xcats, _cids][::flipper][0])
                ]

                if tks == [compare_series]:
                    continue

                plot_dict[i] = {
                    "X": "real_date",
                    "Y": tks + ([compare_series] if compare_series else []),
                    "title": facet_titles[i],
                }
                # plot_dict[i] --> Dict[str, Union[str, List[str]]]

        if cid_xcat_grid:
            # NB : legend goes away in cid_xcat_grid
            legend: bool = False

            for i, cid in enumerate(_cids):
                for j, xcat in enumerate(_xcats):
                    tk: str = "_".join([cid, xcat])
                    plot_dict[i * len(_xcats) + j] = {
                        "X": "real_date",
                        "Y": [tk],
                        "title": tk,
                    }
                    # plot_dict[i * len(_xcats) + j] --> Dict[str, List[str]]

        if not any([cid_grid, xcat_grid, cid_xcat_grid]):
            for i, (key, plt_dct) in enumerate(plot_dict.items()):
                plt_dct["title"] = plt_dct["Y"][0] + " vs " + plt_dct["X"][0]

        if len(plot_dict) == 0:
            raise ValueError("Unable to resolve plot settings.")

        # sort by the title - only
        if all("title" in ditem.keys() for ditem in plot_dict.values()):
            _plot_dict: Dict[str, Dict[str, Union[str, List[str]]]] = {
                i: ditem for i, ditem in enumerate(plot_dict.values())
            }
            plot_dict: Dict[str, Dict[str, Union[str, List[str]]]] = _plot_dict.copy()

        ##############################
        # Plotting
        ##############################

        fig = plt.figure(figsize=figsize)

        outer_gs: GridSpec = GridSpec(
            ncols=grid_dim[0],
            nrows=grid_dim[1],
            figure=fig,
        )

        # re_adj: List[float] = [Left, Bottom, Right, Top]
        re_adj: List[float] = [0, 0, 1, 1]

        suptitle = fig.suptitle(
            title,
            fontsize=title_fontsize,
            x=title_xadjust,
            y=title_yadjust,
        )

        if title is not None:
            # get the figure coordinates of the title
            fig_width, fig_height = (
                suptitle.get_window_extent().width,
                suptitle.get_window_extent().height,
            )
            title_newline_adjust: float = title_fontsize / 500
            if title_yadjust is not None and title_yadjust != 1.0:
                title_newline_adjust = abs(title_yadjust - 1.0)
            # count the number of newlines in the title

            title_height: float = (title.count("\n") + 1) * title_newline_adjust

            re_adj[3] = (
                re_adj[3] - title_height + fig_height / fig.get_window_extent().height
            )

        axs: Union[np.ndarray, plt.Axes] = outer_gs.subplots(
            sharex=share_x,
            sharey=share_y,
        )

        if not isinstance(axs, np.ndarray):
            axs = np.array([axs])
        ax_list = axs.flatten().tolist()

        self.df.set_index(["cid", "xcat"], inplace=True)
        self.df.sort_index(inplace=True)

        for i, (key, plt_dct) in enumerate(plot_dict.items()):
            ax_i = ax_list[i]

            if plt_dct["X"] != "real_date":
                raise NotImplementedError(
                    "Only `real_date` is supported for the X axis."
                )

            is_empty_plot = False

            for iy, y in enumerate(plt_dct["Y"]):
                cidx, xcatx = str(y).split("_", 1)
                try:
                    selected_df = self.df.loc[cidx, xcatx]
                    is_valid_series = True
                except KeyError:
                    is_valid_series = False
                is_empty_plot = is_empty_plot and not is_valid_series

                if is_valid_series:
                    X = selected_df[plt_dct["X"]].values
                    Y = selected_df[metric].values
                else:
                    X, Y = [], []

                plot_func_args = {}
                if legend_color_map:
                    plot_func_args["color"] = legend_color_map.get(
                        xcatx if cid_grid else cidx
                    )

                if y == compare_series:
                    plot_func_args["color"] = "red"
                    plot_func_args["linestyle"] = "--"

                if not interpolate:
                    X, Y = self._insert_nans(X, Y)
                ax_i.plot(X, Y, **plot_func_args)

            if not cid_xcat_grid:
                if facet_titles:
                    ax_i.set_title(
                        plt_dct["title"],
                        fontsize=facet_title_fontsize,
                        x=facet_title_xadjust,
                        y=facet_title_yadjust,
                    )
                if x_axis_label is not None:
                    ax_i.set_xlabel(x_axis_label, fontsize=axis_fontsize)
                if y_axis_label is not None:
                    ax_i.set_ylabel(y_axis_label, fontsize=axis_fontsize)
            else:
                if i < grid_dim[0]:
                    ax_i.set_title(
                        plt_dct["Y"][0].split("_", 1)[1],
                        fontsize=axis_fontsize,
                    )
                if i % grid_dim[0] == 0:
                    ax_i.set_ylabel(
                        plt_dct["Y"][0].split("_", 1)[0],
                        fontsize=axis_fontsize,
                    )

            if is_empty_plot:
                ax_i.set_xticks([])
                ax_i.set_yticks([])
            else:
                if ax_grid:
                    ax_i.grid(axis="both", linestyle="--", alpha=0.5)
                if ax_hline is not None:
                    ax_i.axhline(ax_hline, color="black", linestyle="--")
                if ax_vline is not None:
                    raise NotImplementedError(
                        "Vertical axis lines are not supported at this time."
                    )

        self.df.reset_index(inplace=True)

        for ax in ax_list[len(plot_dict) :]:
            fig.delaxes(ax)
        ax_list = ax_list[: len(plot_dict)]

        if share_x:
            for target_ax in ax_list[(len(plot_dict) - grid_dim[0]) :]:
                target_ax.xaxis.set_tick_params(labelbottom=True)

        if legend:
            if "lower" in legend_loc and legend_ncol < grid_dim[0]:
                legend_ncol = grid_dim[0]

            leg = fig.legend(
                labels=legend_labels,
                loc=legend_loc,
                fontsize=legend_fontsize,
                ncol=legend_ncol,
                bbox_to_anchor=legend_bbox_to_anchor,
                frameon=legend_frame,
            )

            leg_width, leg_height = (
                leg.get_window_extent().width,
                leg.get_window_extent().height,
            )
            fig_width, fig_height = (
                fig.get_window_extent().width,
                fig.get_window_extent().height,
            )

            if "lower" in legend_loc:
                re_adj[1] += leg_height / fig_height
            if "upper" in legend_loc:
                re_adj[3] -= leg_height / fig_height
            if "left" in legend_loc:
                re_adj[0] += leg_width / fig_width
            if "right" in legend_loc:
                re_adj[2] -= leg_width / fig_width

        outer_gs.tight_layout(fig, rect=re_adj)

        if save_to_file is not None:
            fig.savefig(save_to_file, dpi=dpi)

        if show:
            plt.show()

        if return_figure:
            return fig

    def _insert_nans(self, X: Union[np.ndarray, List], Y: Union[np.ndarray, List]):
        """
        Inserts a single NaN value in each inferred gap in the time series.

        :param <np.ndarray> X: Array of dates.
        :param <np.ndarray> Y: Array of values.

        :return <Tuple[np.ndarray, np.ndarray]>: Tuple of arrays with gaps filled.
        """

        if len(X) == 0 or len(Y) == 0:
            return X, Y

        df = pd.DataFrame({"X": X, "Y": Y})
        df["X"] = pd.to_datetime(df["X"])

        # Infer the most common frequency of the time series
        differences = df["X"].diff().dt.total_seconds().dropna()
        frequency_seconds = differences.mode()[0]
        frequency = pd.to_timedelta(frequency_seconds, unit="s")

        # Ignore weekends if daily
        freq_threshold = max(frequency_seconds * 1.5, 60 * 60 * 24 * 3)
        gaps = differences > freq_threshold
        gap_positions = gaps[gaps].index.tolist()

        new_rows = []
        for pos in gap_positions:
            gap_date = df.loc[pos - 1, "X"] + frequency
            new_rows.append({"X": gap_date, "Y": np.nan})

        if new_rows:
            new_rows_df = pd.DataFrame(new_rows)
            df = (
                pd.concat([df, new_rows_df], ignore_index=True)
                .sort_values(by="X")
                .reset_index(drop=True)
            )

        return df.X, df.Y


if __name__ == "__main__":
    # from macrosynergy.visuals import FacetPlot
    from macrosynergy.management.simulate import make_test_df

    import time

    np.random.seed(0)

    cids_A: List[str] = ["AUD", "CAD", "EUR", "GBP", "USD"]
    cids_B: List[str] = ["CHF", "INR", "JPY", "NOK", "NZD", "SEK"]
    cids_C: List[str] = ["CHF", "EUR", "INR", "JPY", "NOK", "NZD", "SEK", "USD"]

    xcats_A: List[str] = [
        "CPIXFE_SA_P1M1ML12",
        "CPIXFE_SJA_P3M3ML3AR",
        "CPIXFE_SJA_P6M6ML6AR",
        "CPIXFE_SA_P1M1ML12_D1M1ML3",
        "CPIXFE_SA_P1M1ML12_D1M1ML3",
    ]
    xcats_B: List[str] = [
        "CPIC_SA_P1M1ML12",
        "CPIC_SJA_P3M3ML3AR",
        "CPIC_SJA_P6M6ML6AR",
        "CPIC_SA_P1M1ML12_D1M1ML3",
        "CPIH_SA_P1M1ML12",
        "EXALLOPENNESS_NSA_1YMA",
        "EXMOPENNESS_NSA_1YMA",
    ]
    xcats_C: List[str] = ["DU05YXR_NSA", "DU05YXR_VT10"]
    xcats_D: List[str] = [
        "FXXR_NSA",
        "EQXR_NSA",
        "FXTARGETED_NSA",
        "FXUNTRADABLE_NSA",
    ]
    all_cids: List[str] = sorted(list(set(cids_A + cids_B + cids_C)))
    all_xcats: List[str] = sorted(list(set(xcats_A + xcats_B + xcats_C + xcats_D)))

    df: pd.DataFrame = make_test_df(
        cids=all_cids,
        xcats=all_xcats,
    )

    # remove data for USD_FXXR_NSA and CHF _EQXR_NSA and _FXXR_NSA
    df: pd.DataFrame = df[
        ~((df["cid"] == "USD") & (df["xcat"] == "FXXR_NSA"))
    ].reset_index(drop=True)
    df: pd.DataFrame = df[
        ~((df["cid"] == "CHF") & (df["xcat"].isin(["EQXR_NSA", "FXXR_NSA"])))
    ].reset_index(drop=True)
    df: pd.DataFrame = df[
        ~((df["cid"] == "NOK") & (df["xcat"] == "FXUNTRADABLE_NSA"))
    ].reset_index(drop=True)

    timer_start: float = time.time()

    with FacetPlot(
        df,
    ) as fp:
        fp.lineplot(
            cids=cids_A,
            share_x=True,
            xcat_grid=True,
            ncols=2,
            title=(
                "Test Title with a very long title to see how it looks, \n and a "
                "new line - why not?"
            ),
            # save_to_file="test_0.png",
            ax_hline=75,
            show=True,
        )

        fp.lineplot(
            cids=cids_B,
            xcats=xcats_A,
            attempt_square=True,
            share_y=True,
            cid_grid=True,
            title="Another test title",
            # save_to_file="test_1.png",
            show=True,
            share_x=True,
        )
        fp.lineplot(
            cids=cids_C,
            xcats=xcats_D,
            cid_xcat_grid=True,
            title="Another test title",
            show=True,
        )

    print(f"Time taken: {time.time() - timer_start}")



"""
Implementation of linear_composite() function as a module.
"""

import numpy as np
import pandas as pd
from typing import List, Dict, Union, Optional, Tuple, Type
import warnings

from macrosynergy.management.utils import reduce_df
from macrosynergy.management.simulate import make_test_df
from macrosynergy.management.utils import is_valid_iso_date

listtypes: Tuple[Type, ...] = (list, np.ndarray, pd.Series, tuple)


def _linear_composite_basic(
    data_df: pd.DataFrame,
    weights_df: pd.DataFrame,
    normalize_weights: bool = True,
    complete: bool = False,
    mode: str = "xcat_agg",
):
    """Main calculation function for linear_composite()"""

    # Create a boolean mask to help us work out the calcs
    nan_mask: pd.DataFrame = data_df.isna() | weights_df.isna()

    # Normalize weights (if requested)
    if normalize_weights:
        adj_weights_wide = weights_df[~nan_mask].div(
            weights_df[~nan_mask].abs().sum(axis=1), axis=0
        )
        adj_weights_wide[nan_mask] = np.NaN

        assert np.allclose(
            adj_weights_wide[~adj_weights_wide.isna().all(axis=1)].abs().sum(axis=1), 1
        ), "Weights do not sum to 1. Normalization failed."

        weights_df = adj_weights_wide.copy()

    # Multiply the weights by the target data
    out_df = data_df * weights_df

    # Sum across the columns
    out_df = out_df.sum(axis="columns")

    # NOTE: Using `axis` with strings, to make it more readable
    # Remove periods with missing data (if requested) (rows with any NaNs)
    if complete:
        out_df[nan_mask.any(axis="columns")] = np.NaN

    # put NaNs back in, as sum() removes them
    out_df[nan_mask.all(axis="columns")] = np.NaN

    # Reset index, rename columns and return
    out_df = out_df.reset_index().rename(columns={0: "value"})

    # TODO: out_df from cid_agg and xcat_agg are not in the same format...

    return out_df


def linear_composite_cid_agg(
    df: pd.DataFrame,
    xcat: str,
    cids: List[str],
    weights: Union[str, List[float]],
    signs: List[float],
    normalize_weights: bool = True,
    complete_cids: bool = True,
    new_cid="GLB",
):
    """Linear composite of various cids for a given xcat across all periods."""

    if isinstance(weights, str):
        weights_df: pd.DataFrame = df[(df["xcat"] == weights)].copy()
        df = df[(df["xcat"] != weights)].copy()
        weights_df = weights_df.set_index(["real_date", "cid"])["value"].unstack(
            level=1
        )
        weights_df = weights_df.mul(signs, axis=1)

    else:
        weights_series: pd.Series = pd.Series(
            np.array(weights) * np.array(signs),
            index=cids,
        )
        weights_df = pd.DataFrame(
            data=[weights_series.sort_index()],
            index=pd.to_datetime(df["real_date"].unique().tolist()),
            columns=df["cid"].unique(),
        )

        weights_df.index.names = ["real_date"]
        weights_df.columns.names = ["cid"]

    # create the data_df
    data_df: pd.DataFrame = (
        df[(df["xcat"] == xcat)]
        .set_index(["real_date", "cid"])["value"]
        .unstack(level=1)
    )
    # aligning the index of weights_df to the data one
    # so that we have the same set of dates and same set of CIDs -- thank you
    # @mikiinterfiore
    weights_df = (
        weights_df.stack(dropna=False)
        .reindex(data_df.stack(dropna=False).index)
        .unstack(1)
    )

    # assert that data_df and weights_df have the same shape, index and columns
    assert (
        (data_df.shape == weights_df.shape)
        and (data_df.index.equals(weights_df.index))
        and (data_df.columns.equals(weights_df.columns))
    ), (
        "Unexpected shape of `data_df` and `weights_df`. "
        "Unable to shape data for calculation."
    )

    # Calculate the linear combination
    out_df: pd.DataFrame = _linear_composite_basic(
        data_df=data_df,
        weights_df=weights_df,
        normalize_weights=normalize_weights,
        complete=complete_cids,
        mode="cid_agg",
    )
    out_df["cid"] = new_cid
    out_df["xcat"] = xcat
    out_df = out_df[["cid", "xcat", "real_date", "value"]]
    return out_df


def linear_composite_xcat_agg(
    df: pd.DataFrame,
    xcats: List[str],
    weights: List[float],
    signs: List[float],
    normalize_weights: bool = True,
    complete_xcats: bool = True,
    new_xcat="NEW",
):
    """Linear composite of various xcats across all cids and periods"""

    # Create a weights series with the xcats as index
    weights_series: pd.Series = pd.Series(
        np.array(weights) * np.array(signs), index=xcats
    )

    # Create wide dataframes for the data and weights
    data_df = df.set_index(["cid", "real_date", "xcat"])["value"].unstack(level=2)
    weights_df = pd.DataFrame(
        data=[weights_series.sort_index()],
        index=data_df.index,
        columns=data_df.columns,
    )

    # Calculate the linear combination
    out_df: pd.DataFrame = _linear_composite_basic(
        data_df=data_df,
        weights_df=weights_df,
        normalize_weights=normalize_weights,
        complete=complete_xcats,
        mode="xcat_agg",
    )
    out_df["xcat"] = new_xcat
    out_df = out_df[["cid", "xcat", "real_date", "value"]]
    return out_df


def linear_composite(
    df: pd.DataFrame,
    xcats: Union[str, List[str]],
    cids: Optional[List[str]] = None,
    weights: Optional[Union[List[float], str]] = None,
    normalize_weights: bool = True,
    signs: Optional[List[float]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    blacklist: Dict[str, List[str]] = None,
    complete_xcats: bool = False,
    complete_cids: bool = False,
    new_xcat="NEW",
    new_cid="GLB",
):
    """
    Weighted linear combinations of cross sections or categories

    :param <pd.DataFrame> df:  standardized JPMaQS DataFrame with the necessary
        columns: 'cid', 'xcat', 'real_date' and 'value'.
    :param <Union[str, List[str]> xcats: One or more categories to be combined.
        If a single category is given the linear combination is calculated across
        sections. This results in a single series to which a new cross-sectional
        identifier is assigned.
        If more than one category string is given the output will be a new category,
        i.e. a panel that is a linear combination of the categories specified.
    :param <List[str]> cids: cross-sections for which the linear combinations are
        calculated. Default is all cross-section available.
    :param <Union[List[float], str]> weights: This specifies how categories or cross
        sections are combined. There are three principal options.
        The first (default) is None, in which case equal weights are given to all
        categories or cross sections that are available.
        The second case is a set of fixed coefficients, in which case these very
        coefficients are applied to all available categories of cross sections.
        Per default the coefficients are normalized so that they add up to one for each
        period. This can be changed with the argument `normalize_weights`.
        The third case is the assignment of a weighting category. This only applies to
        combinations of cross sections. In this care the weighting category is multiplied
        for each period with the corresponding value of main category of the same cross
        section. Per default the weight category values are normalized so that they add up
        to one for each period. This can be changed with the argument `normalize_weights`.
    :param <bool> normalize_weights: If True (default) the weights are normalized to sum
        to 1. If False the weights are used as specified.
    :param <List[float]> signs: An array of consisting of +1s or -1s, of the same length
        as the number of categories in `xcats` to indicate whether the respective category
        should be added or subtracted from the linear combination. Not relevant when
        aggregating over cross-sections, i.e. when a single category is given in `xcats`.
        Default is None and all signs are set to +1.
    :param <str> start: earliest date in ISO format. Default is None and earliest date
        for which the respective category is available is used.
    :param <str> end: latest date in ISO format. Default is None and latest date for
        which the respective category is available is used.
    :param <bool> complete_xcats: If True (default) combinations are only calculated for
        observation dates on which all xcats are available. If False a combination of the
        available categories is used. Not relevant when aggregating over cross-sections,
        i.e. when a single category is given in `xcats`.
    :param <bool> complete_cids: If True (default) combinations are only calculated for
        observation dates on which all cids are available. If False a combination of the
        available cross-sections is used. Not relevant when aggregating over categories,
        i.e. when multiple categories are given in `xcats`.
    :param <str> new_xcat: Name of new composite xcat when aggregating over xcats for a
        given cid. Default is "NEW".
    :param <str> new_cid: Name of new composite cid when aggregating over cids for a given
        xcat. Default is "GLB".

    :return <pd.DataFrame>: standardized DataFrame with the relative values, featuring
        the categories: 'cid', 'xcat', 'real_date' and 'value'.
    """

    # df check
    if not isinstance(df, pd.DataFrame):
        raise TypeError("`df` must be a pandas DataFrame")

    if not set(["cid", "xcat", "real_date", "value"]).issubset(set(df.columns)):
        raise ValueError(
            "`df` must be a standardized JPMaQS DataFrame with the necessary columns: "
            "'cid', 'xcat', 'real_date' and 'value'."
        )

    if df["value"].isna().all():
        raise ValueError("`df` does not contain any valid values.")

    # copy df to avoid side effects
    # NOTE: The below arg validation contains code that "copy" the args.
    # Be careful when making changes.
    df: pd.DataFrame = df.copy()

    if start is None:
        start: str = pd.to_datetime(df["real_date"]).min().strftime("%Y-%m-%d")
    if end is None:
        end: str = pd.to_datetime(df["real_date"]).max().strftime("%Y-%m-%d")

    # dates check
    for varx, namex in zip([start, end], ["start", "end"]):
        if varx is not None:
            if not (isinstance(varx, str) and is_valid_iso_date(varx)):
                raise ValueError(f"`{namex}` must be a valid ISO date string.")

    # check xcats
    if xcats is None:
        xcats: List[str] = df["xcat"].unique().tolist()
    elif isinstance(xcats, str):
        xcats: List[str] = [xcats]
    elif isinstance(xcats, listtypes):
        xcats: List[str] = list(xcats)
    else:
        raise TypeError("`xcats` must be a string or list of strings.")

    # check xcats in df
    if not set(xcats).issubset(set(df["xcat"].unique().tolist())):
        raise ValueError("Not all `xcats` are available in `df`.")

    # check cids
    if cids is None:
        cids: List[str] = df["cid"].unique().tolist()
    elif isinstance(cids, str):
        cids: List[str] = [cids]
    elif isinstance(cids, listtypes):
        cids: List[str] = list(cids)
    else:
        raise TypeError("`cids` must be a string or list of strings.")

    # check cids in df
    if not set(cids).issubset(set(df["cid"].unique().tolist())):
        raise ValueError("Not all `cids` are available in `df`.")

    _xcat_agg: bool = len(xcats) > 1 or new_xcat != "NEW"
    mode: str = "xcat_agg" if _xcat_agg else "cid_agg"

    if _xcat_agg and isinstance(weights, str):
        raise ValueError(
            "When aggregating over xcats, `weights` "
            "must be a list of floats or integers."
        )

    # check weights
    expc_weights_len: int = len(xcats) if _xcat_agg else len(cids)

    if weights is None:
        weights: List[float] = list(np.ones(expc_weights_len) / expc_weights_len)
    elif isinstance(weights, listtypes):
        weights: List[float] = list(weights)
        if not all([isinstance(x, (float, int)) for x in weights]):
            raise TypeError("`weights` must be a list of floats or integers.")
        if len(weights) != expc_weights_len:
            raise ValueError(
                "`weights` must be a list of floats of the same length as `xcats`."
            )
        if any([x == 0.0 for x in weights]):
            raise ValueError("`weights` must not contain any 0s.")

    elif isinstance(weights, str):
        if weights not in df["xcat"].unique().tolist():
            raise ValueError(
                "When using a category-string as `weights`"
                " it must be present in `df`."
            )
    else:
        raise TypeError("`weights` must be a list of floats, a string or None.")

    # check signs
    if signs is None:
        signs: List[float] = [1.0] * (len(xcats) if _xcat_agg else len(cids))
    elif isinstance(signs, listtypes):
        signs: List[float] = list(signs)
        if len(signs) != expc_weights_len:
            raise ValueError(
                "`signs` must be a list of floats of the same length as `xcats`."
            )
        if not all([x in [-1.0, 1.0] for x in signs]):
            if any([x == 0.0 for x in signs]):
                raise ValueError("`signs` must not contain any 0s.")
            warnings.warn(
                "`signs` must be a list of +1s or -1s. "
                "`signs` will be coerced to +1s/-1s. "
                "(i.e. signs ← abs(signs) / signs)"
            )

            signs: List[float] = [abs(x) / x for x in signs]

    else:
        raise TypeError("`signs` must be a list of floats/ints or None.")

    _xcats: List[str] = xcats + ([weights] if isinstance(weights, str) else [])

    df: pd.DataFrame
    remaining_xcats: List[str]
    remaining_cids: List[str]
    # NOTE: the "remaining_*" variables will not be in the same order as the input
    # cids/xcats.
    # Do not used these for index based lookups/operations.
    df, remaining_xcats, remaining_cids = reduce_df(
        df=df,
        xcats=_xcats,
        cids=cids,
        start=start,
        end=end,
        blacklist=blacklist,
        intersect=False,
        out_all=True,
    )
    if len(remaining_xcats) == 1 and len(remaining_cids) < len(cids) and not _xcat_agg:
        raise ValueError(
            "Not all `cids` have complete `xcat` data required for the calculation."
        )

    if _xcat_agg:
        found_cids: List[str] = df["cid"].unique().tolist()
        found_xcats: List[str] = df["xcat"].unique().tolist()

        for icid, cidx in enumerate(found_cids):
            missing_xcats: List[str] = list(
                set(found_xcats) - set(df.loc[df["cid"] == cidx, "xcat"].unique())
            )
            if missing_xcats:
                # warn the user, and put in the dates with NaNs
                warnings.warn(
                    f"`cid` {cidx} does not have complete `xcat` data for "
                    f"{missing_xcats}."
                    " These will be filled with NaNs for the calculation."
                )
                # artificially add the missing xcats
                dt_range: pd.DatetimeIndex = pd.to_datetime(df["real_date"].unique())
                for xc in missing_xcats:
                    df = pd.concat(
                        [
                            df,
                            pd.DataFrame(
                                data={
                                    "cid": cidx,
                                    "xcat": xc,
                                    "real_date": dt_range,
                                    "value": np.NaN,
                                }
                            ),
                        ]
                    )

        return linear_composite_xcat_agg(
            df=df,
            xcats=xcats,
            weights=weights,
            signs=signs,
            normalize_weights=normalize_weights,
            complete_xcats=complete_xcats,
            new_xcat=new_xcat,
        )

    else:  # mode == "cid_agg" -- single xcat
        found_cids: List[str] = df["cid"].unique().tolist()
        found_xcats: List[str] = df["xcat"].unique().tolist()
        if isinstance(weights, str):
            # one of the found_xcats must be the weights, and there should be only one
            # more
            assert (weights in found_xcats) and len(
                (set(found_xcats) - {weights})
            ) == 1, (
                "When using a category-string as `weights`"
                " it must be present in `df` and there must be only one other `xcat`."
            )

        ctr = 0
        for cidx in cids.copy():  # copy to allow modification of `cids`
            missing_xcats: List[str] = list(
                set(found_xcats) - set(df.loc[df["cid"] == cidx, "xcat"].unique())
            )
            if missing_xcats:
                cids.pop(ctr)
                signs.pop(ctr)
                if isinstance(weights, list):
                    weights.pop(ctr)
                # drop from df
                df = df.loc[df["cid"] != cidx, :]
                warnings.warn(
                    f"`cid` {cidx} does not have complete `xcat` data for "
                    f"{missing_xcats}."
                    " It will be dropped from dataframe."
                )
            else:
                ctr += 1

        if len(cids) == 0:
            raise ValueError(
                "No `cids` have complete `xcat` data required for the calculation."
            )

        _xcat: str = list(
            set(found_xcats) - {weights if isinstance(weights, str) else ""}
        )[0]

        return linear_composite_cid_agg(
            df=df,
            xcat=_xcat,
            cids=cids,
            weights=weights,
            signs=signs,
            normalize_weights=normalize_weights,
            complete_cids=complete_cids,
            new_cid=new_cid,
        )


if __name__ == "__main__":
    cids = ["AUD", "CAD", "GBP"]
    xcats = ["XR", "CRY", "INFL"]

    df: pd.DataFrame = pd.concat(
        [
            make_test_df(
                cids=cids,
                xcats=xcats[:-1],
                start="2000-01-01",
                end="2000-02-01",
                style="linear",
            ),
            make_test_df(
                cids=cids,
                xcats=["INFL"],
                start="2000-01-01",
                end="2000-02-01",
                style="decreasing-linear",
            ),
        ]
    )

    # all infls are now decreasing-linear, while everything else is increasing-linear

    df.loc[
        (df["cid"] == "GBP")
        & (df["xcat"] == "INFL")
        & (df["real_date"] == "2000-01-17"),
        "value",
    ] = np.NaN

    df.loc[
        (df["cid"] == "AUD")
        & (df["xcat"] == "CRY")
        & (df["real_date"] == "2000-01-17"),
        "value",
    ] = np.NaN

    # there are now missing values for AUD-CRY and GBP-INFL on 2000-01-17

    lc_cid = linear_composite(
        df=df, xcats="XR", weights="INFL", normalize_weights=False
    )

    lc_xcat = linear_composite(
        df=df,
        cids=["AUD", "CAD"],
        xcats=["XR", "CRY", "INFL"],
        weights=[1, 2, 1],
        signs=[1, -1, 1],
    )



import numpy as np
from sklearn.base import BaseEstimator, RegressorMixin


class NaivePredictor(BaseEstimator, RegressorMixin):
    """
    Naive predictor class to output a column of features as the predictions themselves.
    It expects only one column in the input matrix for which predictions are to be made.
    This class would typically be used in the final stage of a `scikit-learn` pipeline,
    following either `FeatureAverager` or `ZnScoreAverager` stages.
    """

    def fit(self, X, y=None):
        return self

    def predict(self, X):
        return np.array(X)



"""
A subclass inheriting from `macrosynergy.visuals.plotter.Plotter`,
designed to plot time series data on a line plot.
"""

import pandas as pd
from typing import List, Dict, Tuple, Optional, Union
import matplotlib.pyplot as plt
from macrosynergy.management.utils import is_valid_iso_date
from macrosynergy.visuals.plotter import Plotter
from numbers import Number
from macrosynergy.management.simulate import make_test_df


class LinePlot(Plotter):
    """
    Class for plotting time series data on a line plot.
    Inherits from `macrosynergy.visuals.plotter.Plotter`.

    Parameters
    :param <pd.DataFrame> df: A DataFrame with the following columns:
        'cid', 'xcat', 'real_date', and at least one metric from -
        'value', 'grading', 'eop_lag', or 'mop_lag'.
    :param <List[str]> cids: A list of cids to select from the DataFrame.
        If None, all cids are selected.
    :param <List[str]> xcats: A list of xcats to select from the DataFrame.
        If None, all xcats are selected.
    :param <List[str]> metrics: A list of metrics to select from the DataFrame.
        If None, all metrics are selected.
    :param <str> start_date: ISO-8601 formatted date. Select data from
        this date onwards. If None, all dates are selected.
    :param <str> end_date: ISO-8601 formatted date. Select data up to
        and including this date. If None, all dates are selected.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        cids: Optional[List[str]] = None,
        xcats: Optional[List[str]] = None,
        metrics: Optional[List[str]] = None,
        intersect: bool = False,
        tickers: Optional[List[str]] = None,
        blacklist: Optional[Dict[str, List[str]]] = None,
        start: Optional[str] = None,
        end: Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(
            df=df,
            cids=cids,
            xcats=xcats,
            metrics=metrics,
            intersect=intersect,
            tickers=tickers,
            blacklist=blacklist,
            start=start,
            end=end,
            *args,
            **kwargs,
        )

    def plot(
        self,
        # df/plot args
        metric: str = "value",
        compare_series: Optional[str] = None,
        # Plotting specific arguments
        # fig args
        figsize: Tuple[Number, Number] = (12, 8),
        aspect: Number = 1.618,
        height: Number = 0.8,
        # plot args
        grid: bool = True,
        x_axis_label: Optional[str] = None,
        y_axis_label: Optional[str] = None,
        axis_fontsize: int = 12,
        ax_hline: Optional[Union[Number, List[Number]]] = None,
        ax_vline: Optional[Union[str, List[str]]] = None,
        # title args
        title: Optional[str] = None,
        title_fontsize: int = 16,
        title_xadjust: Number = 0.5,
        title_yadjust: Number = 1.05,
        # legend args
        legend: bool = True,
        legend_labels: Optional[List[str]] = None,
        legend_title: Optional[str] = None,
        legend_loc: Optional[str] = "best",
        legend_fontsize: int = 14,
        legend_ncol: int = 1,
        legend_bbox_to_anchor: Optional[Tuple[Number, Number]] = None,
        legend_frame: bool = True,
        # return args
        show: bool = True,
        save_to_file: Optional[str] = None,
        dpi: int = 300,
        return_figure: bool = False,
        on_axis: Optional[plt.Axes] = None,
        # args, kwargs
        *args,
        **kwargs,
    ):
        if on_axis:
            fig: plt.Figure = on_axis.get_figure()
            ax: plt.Axes = on_axis
        else:
            fig: plt.Figure
            ax: plt.Axes
            fig, ax = plt.subplots(figsize=figsize)

        dfx: pd.DataFrame = self.df.copy()

        if ax_vline is not None:
            # check it's all valid iso-8601 dates
            if isinstance(ax_vline, str):
                ax_vline = [ax_vline]
            for vline in ax_vline:
                if not is_valid_iso_date(vline):
                    raise ValueError(
                        f"Invalid date format for `ax_vline`: {vline}. "
                        "Please use ISO-8601 formatted dates."
                    )

        if ax_hline is not None:
            if isinstance(ax_hline, Number):
                ax_hline = [ax_hline]
            for hline in ax_hline:
                if not isinstance(hline, Number):
                    raise ValueError(
                        f"Invalid value for `ax_hline`: {hline}. "
                        "Please use a number."
                    )

        if title is not None:
            title_newline_adjust: float = 0.2
            if title_yadjust is not None and title_yadjust != 1.0:
                title_newline_adjust = abs(title_yadjust - 1.0)
            title_height: float = title.count("\n") * title_newline_adjust
            # increase the title_yadjust by the title_height
            title_yadjust = title_yadjust + title_height

        ax.set_title(
            title,
            fontsize=title_fontsize,
            x=title_xadjust,
            y=title_yadjust,
        )

        if compare_series:
            _cid, _xcat = compare_series.split("_", 1)
            if _cid not in dfx["cid"].unique() or _xcat not in dfx["xcat"].unique():
                raise ValueError(
                    f"Series `{compare_series}` not in DataFrame - used as "
                    "`compare_series`."
                )

            comp_df = (
                dfx.loc[
                    (dfx["cid"] == _cid) & (dfx["xcat"] == _xcat), ["real_date", metric]
                ]
                .copy()
                .reset_index(drop=True)
            )

        valid_tickers = dfx[["cid", "xcat"]].drop_duplicates().values.tolist()

        for xcat in self.xcats:
            for cid in self.cids:
                # Get the unique cid values in dfx for xcat and check if cid is in it
                if [cid, xcat] in valid_tickers:
                    _df = dfx.loc[(dfx["cid"] == cid) & (dfx["xcat"] == xcat), :].copy()
                    _df = _df.sort_values(by="real_date", ascending=True).reset_index(
                        drop=True
                    )
                    ax.plot(_df["real_date"], _df[metric], label=f"{cid}_{xcat}")

        # if there is a compare_series, plot it on the same axis, using a red dashed line
        if compare_series:
            ax.plot(
                comp_df["real_date"],
                comp_df[metric],
                color="red",
                linestyle="--",
                label=compare_series,
            )

        if ax_vline is not None:
            for vline in ax_vline:
                ax.axvline(x=pd.to_datetime(vline), color="black", linestyle="--")

        if ax_hline is not None:
            for hline in ax_hline:
                ax.axhline(y=hline, color="black", linestyle="--")

        if grid:
            ax.grid(axis="both", linestyle="--", alpha=0.5)

        if x_axis_label:
            ax.set_xlabel(x_axis_label, fontsize=axis_fontsize)

        if y_axis_label:
            ax.set_ylabel(y_axis_label, fontsize=axis_fontsize)

        # if there is a legend, add it
        if legend:
            ax.legend(
                labels=legend_labels if legend_labels else None,
                title=legend_title,
                loc=legend_loc,
                fontsize=legend_fontsize,
                ncol=legend_ncol,
                bbox_to_anchor=legend_bbox_to_anchor,
                frameon=legend_frame,
            )

        plt.tight_layout()
        title: str = title if title else f"LinePlot: Viewing `{metric}`"

        if save_to_file:
            plt.savefig(
                save_to_file,
                dpi=dpi,
                bbox_inches="tight",
            )

        if return_figure:
            return fig

        if show:
            plt.show()
            return


if __name__ == "__main__":
    from macrosynergy.management.simulate import make_test_df

    cids = ["USD", "EUR", "GBP", "JPY"]
    xcats = ["RIR_NSA", "FXXR_NSA", "EQXR_NSA"]
    df = make_test_df(cids=cids, xcats=xcats)

    import time

    timer_start: float = time.time()

    LinePlot(df).plot(
        title=(
            "Test Title with a very long title to see how it looks, \n and a new "
            "line - why not?"
        ),
        legend_fontsize=8,
        compare_series="USD_RIR_NSA",
        ax_hline=[45, 55],
        ax_vline=["2015-4-01", "2015-09-01"],
    )

    # facet_size=(5, 4),
    print(f"Time taken: {time.time() - timer_start}")
